# Copyright 2023 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from http import HTTPStatus
from typing import AsyncGenerator, List, Union

from fastapi import Request, Response
from starlette.datastructures import Headers

from kserve.protocol.rest.openai.errors import create_error_response
from kserve.protocol.rest.openai.types import (
    ChatCompletion,
    ChatCompletionRequest,
    Completion,
    CompletionRequest,
    Embedding,
    EmbeddingRequest,
    ErrorResponse,
    RerankRequest,
    Rerank,
)
from ...dataplane import DataPlane
from .openai_model import (
    OpenAIModel,
    OpenAIGenerativeModel,
    OpenAIEncoderModel,
)


class OpenAIDataPlane(DataPlane):
    """OpenAI DataPlane"""

    async def create_completion(
        self,
        model_name: str,
        request: CompletionRequest,
        raw_request: Request,
        headers: Headers,
        response: Response,
    ) -> Union[AsyncGenerator[str, None], Completion, ErrorResponse]:
        """Generate the text with the provided text prompt.

        Args:
            model_name (str): Model name.
            request (CompletionRequest): Params to create a completion.
            raw_request (Request): fastapi request object.
            headers: (Headers): Request headers.
            response: (Response): FastAPI response object
        Returns:
            response: A non-streaming or streaming completion response or an error response.
        """
        model = await self.get_model(model_name)
        if not isinstance(model, OpenAIGenerativeModel):
            return create_error_response(
                message=f"Model {model_name} does not support Completions API",
                status_code=HTTPStatus.BAD_REQUEST,
            )

        context = {"headers": dict(headers), "response": response}
        return await model.create_completion(
            request=request, raw_request=raw_request, context=context
        )

    async def create_chat_completion(
        self,
        model_name: str,
        request: ChatCompletionRequest,
        raw_request: Request,
        headers: Headers,
        response: Response,
    ) -> Union[AsyncGenerator[str, None], ChatCompletion, ErrorResponse]:
        """Generate the text with the provided text prompt.

        Args:
            model_name (str): Model name.
            request (CreateChatCompletionRequest): Params to create a chat completion.
            headers: (Optional[Dict[str, str]]): Request headers.

        Returns:
            response: A non-streaming or streaming chat completion response or an error response.
        """
        model = await self.get_model(model_name)
        if not isinstance(model, OpenAIGenerativeModel):
            return create_error_response(
                message=f"Model {model_name} does not support Chat Completion API",
                status_code=HTTPStatus.BAD_REQUEST,
            )

        context = {"headers": dict(headers), "response": response}
        return await model.create_chat_completion(
            request=request, raw_request=raw_request, context=context
        )

    async def create_embedding(
        self,
        model_name: str,
        request: EmbeddingRequest,
        raw_request: Request,
        headers: Headers,
        response: Response,
    ) -> Union[AsyncGenerator[str, None], Embedding, ErrorResponse]:
        """Generate the text with the provided text prompt.

        Args:
            model_name (str): Model name.
            request (EmbeddingRequest): Params to create a embedding.
            raw_request (Request): fastapi request object.
            headers: (Headers): Request headers.
            response: (Response): FastAPI response object
        Returns:
            response: A non-streaming or streaming embedding response or an error response.
        """
        model = await self.get_model(model_name)
        if not isinstance(model, OpenAIEncoderModel):
            return create_error_response(
                message=f"Model {model_name} does not support Embeddings API",
                status_code=HTTPStatus.BAD_REQUEST,
            )

        context = {"headers": dict(headers), "response": response}
        return await model.create_embedding(
            request=request, raw_request=raw_request, context=context
        )

    async def create_rerank(
        self,
        model_name: str,
        request: RerankRequest,
        raw_request: Request,
        headers: Headers,
        response: Response,
    ) -> Union[AsyncGenerator[str, None], Rerank, ErrorResponse]:
        """Generate the text with the provided text prompt.
        Args:
            model_name (str): Model name.
            request (RerankRequest): Params to create rerank response.
            raw_request (Request): fastapi request object.
            headers: (Headers): Request headers.
            response: (Response): FastAPI response object
        Returns:
            Returns:
            response: A non-streaming or streaming embedding response or an error response.
        """
        model = await self.get_model(model_name)
        if not isinstance(model, OpenAIEncoderModel):
            return create_error_response(
                message=f"Model {model_name} does not support Rerank API",
                status_code=HTTPStatus.BAD_REQUEST,
            )

        context = {"headers": dict(headers), "response": response}
        return await model.create_rerank(
            request=request, raw_request=raw_request, context=context
        )

    async def models(self) -> List[OpenAIModel]:
        """Retrieve a list of models

        Returns:
            response: A list of OpenAIModel instances
        """
        return [
            model
            for model in self.model_registry.get_models().values()
            if isinstance(model, OpenAIModel)
        ]



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
KServe

Python SDK for KServe  # noqa: E501

The version of the OpenAPI document: v0.1
Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1alpha1_cluster_serving_runtime_list import (
    V1alpha1ClusterServingRuntimeList,
)  # noqa: E501
from kserve.rest import ApiException


class TestV1alpha1ClusterServingRuntimeList(unittest.TestCase):
    """V1alpha1ClusterServingRuntimeList unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1alpha1ClusterServingRuntimeList
        include_option is a boolean, when False only required
        params are included, when True both required and
        optional params are included"""
        # model = kserve.models.v1alpha1_cluster_serving_runtime_list.V1alpha1ClusterServingRuntimeList()  # noqa: E501
        if include_optional:
            return V1alpha1ClusterServingRuntimeList(
                api_version="0",
                items=[
                    kserve.models.v1alpha1_cluster_serving_runtime.V1alpha1ClusterServingRuntime(
                        api_version="0",
                        kind="0",
                        metadata=None,
                        spec=kserve.models.v1alpha1_serving_runtime_spec.V1alpha1ServingRuntimeSpec(
                            affinity=None,
                            built_in_adapter=kserve.models.v1alpha1_built_in_adapter.V1alpha1BuiltInAdapter(
                                mem_buffer_bytes=56,
                                model_loading_timeout_millis=56,
                                runtime_management_port=56,
                                server_type="0",
                            ),
                            containers=[
                                kserve.models.v1alpha1_container.V1alpha1Container(
                                    args=["0"],
                                    command=["0"],
                                    env=[None],
                                    image="0",
                                    image_pull_policy="0",
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    working_dir="0",
                                )
                            ],
                            disabled=True,
                            grpc_data_endpoint="0",
                            grpc_endpoint="0",
                            http_data_endpoint="0",
                            node_selector={"key": "0"},
                            replicas=56,
                            storage_helper=kserve.models.v1alpha1_storage_helper.V1alpha1StorageHelper(
                                disabled=True,
                            ),
                            supported_model_formats=[
                                kserve.models.v1alpha1_supported_model_format.V1alpha1SupportedModelFormat(
                                    auto_select=True,
                                    name="0",
                                    version="0",
                                )
                            ],
                            tolerations=[None],
                        ),
                        status=None,
                    )
                ],
                kind="0",
                metadata=None,
            )
        else:
            return V1alpha1ClusterServingRuntimeList(
                items=[
                    kserve.models.v1alpha1_cluster_serving_runtime.V1alpha1ClusterServingRuntime(
                        api_version="0",
                        kind="0",
                        metadata=None,
                        spec=kserve.models.v1alpha1_serving_runtime_spec.V1alpha1ServingRuntimeSpec(
                            affinity=None,
                            built_in_adapter=kserve.models.v1alpha1_built_in_adapter.V1alpha1BuiltInAdapter(
                                mem_buffer_bytes=56,
                                model_loading_timeout_millis=56,
                                runtime_management_port=56,
                                server_type="0",
                            ),
                            containers=[
                                kserve.models.v1alpha1_container.V1alpha1Container(
                                    args=["0"],
                                    command=["0"],
                                    env=[None],
                                    image="0",
                                    image_pull_policy="0",
                                    liveness_probe=None,
                                    name="0",
                                    readiness_probe=None,
                                    resources=None,
                                    working_dir="0",
                                )
                            ],
                            disabled=True,
                            grpc_data_endpoint="0",
                            grpc_endpoint="0",
                            http_data_endpoint="0",
                            node_selector={"key": "0"},
                            replicas=56,
                            storage_helper=kserve.models.v1alpha1_storage_helper.V1alpha1StorageHelper(
                                disabled=True,
                            ),
                            supported_model_formats=[
                                kserve.models.v1alpha1_supported_model_format.V1alpha1SupportedModelFormat(
                                    auto_select=True,
                                    name="0",
                                    version="0",
                                )
                            ],
                            tolerations=[None],
                        ),
                        status=None,
                    )
                ],
            )

    def testV1alpha1ClusterServingRuntimeList(self):
        """Test V1alpha1ClusterServingRuntimeList"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == "__main__":
    unittest.main()



# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc_testing
import numpy as np
import pandas as pd
import pytest
from google.protobuf.json_format import MessageToDict
from unittest.mock import patch

from kserve import Model, ModelServer
from kserve.errors import InvalidInput
from kserve.protocol.grpc import grpc_predict_v2_pb2, servicer
from kserve.protocol.infer_type import serialize_byte_tensor, InferResponse
from kserve.utils.utils import get_predict_response


class DummyModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        outputs = pd.DataFrame(
            {
                "fp32_output": request.get_input_by_name("fp32_input")
                .as_numpy()
                .flatten(),
                "int32_output": request.get_input_by_name("int32_input")
                .as_numpy()
                .flatten(),
                "string_output": request.get_input_by_name("string_input")
                .as_numpy()
                .flatten(),
                "uint8_output": request.get_input_by_name("uint8_input")
                .as_numpy()
                .flatten(),
                "bool_input": request.get_input_by_name("bool_input")
                .as_numpy()
                .flatten(),
            }
        )
        # Fixme: Gets only the 1st element of the input
        # inputs = get_predict_input(request)
        infer_response = get_predict_response(request, outputs, self.name)
        if request.parameters:
            infer_response.parameters = request.parameters
        if request.inputs[0].parameters:
            infer_response.outputs[0].parameters = request.inputs[0].parameters
        return infer_response


class DummyFP16OutputModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        outputs = pd.DataFrame(
            {
                "fp16_output": request.get_input_by_name("fp32_input")
                .as_numpy()
                .astype(np.float16)
                .flatten(),
                "fp32_output": request.get_input_by_name("fp32_input")
                .as_numpy()
                .flatten(),
            }
        )
        # Fixme: Gets only the 1st element of the input
        # inputs = get_predict_input(request)
        infer_response = get_predict_response(request, outputs, self.name)
        if request.parameters:
            infer_response.parameters = request.parameters
        if request.inputs[0].parameters:
            infer_response.outputs[0].parameters = request.inputs[0].parameters
        return infer_response


class DummyFP16InputModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        outputs = pd.DataFrame(
            {
                "int32_output": np.array([1, 2, 3, 4, 5, 6, 7, 8]),
                "fp16_output": request.get_input_by_name("fp16_input")
                .as_numpy()
                .flatten(),
            }
        )
        # Fixme: Gets only the 1st element of the input
        # inputs = get_predict_input(request)
        infer_response = get_predict_response(request, outputs, self.name)
        if request.parameters:
            infer_response.parameters = request.parameters
        if request.inputs[0].parameters:
            infer_response.outputs[0].parameters = request.inputs[0].parameters
        return infer_response


@pytest.fixture(scope="class")
def server():
    server = ModelServer()
    model = DummyModel("TestModel")
    model.load()
    server.register_model(model)
    fp16_output_model = DummyFP16OutputModel("FP16OutputModel")
    fp16_output_model.load()
    server.register_model(fp16_output_model)
    fp16_input_model = DummyFP16InputModel("FP16InputModel")
    fp16_input_model.load()
    server.register_model(fp16_input_model)
    servicers = {
        grpc_predict_v2_pb2.DESCRIPTOR.services_by_name[
            "GRPCInferenceService"
        ]: servicer.InferenceServicer(
            server.dataplane, server.model_repository_extension
        )
    }
    test_server = grpc_testing.server_from_dictionary(
        servicers,
        grpc_testing.strict_real_time(),
    )
    return test_server


@pytest.mark.asyncio
@patch(
    "kserve.protocol.grpc.servicer.to_headers", return_value=[]
)  # To avoid NotImplementedError from trailing_metadata function
async def test_grpc_inputs(mock_to_headers, server):
    request = grpc_predict_v2_pb2.ModelInferRequest(
        model_name="TestModel",
        id="123",
        inputs=[
            {
                "name": "fp32_input",
                "shape": [2, 4],
                "datatype": "FP32",
                "contents": {"fp32_contents": [6.8, 2.8, 4.8, 1.4, 6.0, 3.4, 4.5, 1.6]},
            },
            {
                "name": "int32_input",
                "shape": [2, 4],
                "datatype": "INT32",
                "contents": {"int_contents": [6, 2, 4, 1, 6, 3, 4, 1]},
            },
            {
                "name": "string_input",
                "shape": [8],
                "datatype": "BYTES",
                "contents": {
                    "bytes_contents": [
                        b"Cat",
                        b"Dog",
                        b"Wolf",
                        b"Cat",
                        b"Dog",
                        b"Wolf",
                        b"Dog",
                        b"Wolf",
                    ]
                },
            },
            {
                "name": "uint8_input",
                "shape": [2, 4],
                "datatype": "UINT8",
                "contents": {"uint_contents": [6, 2, 4, 1, 6, 3, 4, 1]},
            },
            {
                "name": "bool_input",
                "shape": [8],
                "datatype": "BOOL",
                "contents": {
                    "bool_contents": [
                        True,
                        False,
                        True,
                        False,
                        True,
                        False,
                        True,
                        False,
                    ]
                },
            },
        ],
    )

    model_infer_method = server.invoke_unary_unary(
        method_descriptor=(
            grpc_predict_v2_pb2.DESCRIPTOR.services_by_name[
                "GRPCInferenceService"
            ].methods_by_name["ModelInfer"]
        ),
        invocation_metadata={},
        request=request,
        timeout=20,
    )

    response, _, code, _ = model_infer_method.termination()
    response = await response
    response_dict = MessageToDict(
        response,
        preserving_proto_field_name=True,
    )
    assert code == grpc.StatusCode.OK
    assert response_dict == {
        "model_name": "TestModel",
        "id": "123",
        "outputs": [
            {
                "name": "fp32_output",
                "datatype": "FP32",
                "shape": ["8"],
                "contents": {"fp32_contents": [6.8, 2.8, 4.8, 1.4, 6.0, 3.4, 4.5, 1.6]},
            },
            {
                "name": "int32_output",
                "datatype": "INT32",
                "shape": ["8"],
                "contents": {"int_contents": [6, 2, 4, 1, 6, 3, 4, 1]},
            },
            {
                "name": "string_output",
                "datatype": "BYTES",
                "shape": ["8"],
                "contents": {
                    "bytes_contents": [
                        "Q2F0",
                        "RG9n",
                        "V29sZg==",
                        "Q2F0",
                        "RG9n",
                        "V29sZg==",
                        "RG9n",
                        "V29sZg==",
                    ]
                },
            },
            {
                "name": "uint8_output",
                "datatype": "UINT8",
                "shape": ["8"],
                "contents": {"uint_contents": [6, 2, 4, 1, 6, 3, 4, 1]},
            },
            {
                "name": "bool_input",
                "datatype": "BOOL",
                "shape": ["8"],
                "contents": {
                    "bool_contents": [
                        True,
                        False,
                        True,
                        False,
                        True,
                        False,
                        True,
                        False,
                    ]
                },
            },
        ],
    }


@pytest.mark.asyncio
@patch(
    "kserve.protocol.grpc.servicer.to_headers", return_value=[]
)  # To avoid NotImplementedError from trailing_metadata function
async def test_grpc_raw_inputs(mock_to_headers, server):
    """
    If we receive raw inputs then, the response also should be in raw output format.
    """
    fp32_data = np.array([6.8, 2.8, 4.8, 1.4, 6.0, 3.4, 4.5, 1.6], dtype=np.float32)
    int32_data = np.array([6, 2, 4, 1, 6, 3, 4, 1], dtype=np.int32)
    str_data = np.array(
        [b"Cat", b"Dog", b"Wolf", b"Cat", b"Dog", b"Wolf", b"Dog", b"Wolf"],
        dtype=np.object_,
    )
    uint8_data = np.array([6, 2, 4, 1, 6, 3, 4, 1], dtype=np.uint8)
    bool_data = np.array(
        [True, False, True, False, True, False, True, False], dtype=np.bool_
    )
    raw_input_contents = [
        fp32_data.tobytes(),
        int32_data.tobytes(),
        serialize_byte_tensor(str_data).item(),
        uint8_data.tobytes(),
        bool_data.tobytes(),
    ]
    request = grpc_predict_v2_pb2.ModelInferRequest(
        model_name="TestModel",
        id="123",
        inputs=[
            {
                "name": "fp32_input",
                "shape": [2, 4],
                "datatype": "FP32",
            },
            {
                "name": "int32_input",
                "shape": [2, 4],
                "datatype": "INT32",
            },
            {
                "name": "string_input",
                "shape": [8],
                "datatype": "BYTES",
            },
            {
                "name": "uint8_input",
                "shape": [2, 4],
                "datatype": "UINT8",
            },
            {
                "name": "bool_input",
                "shape": [8],
                "datatype": "BOOL",
            },
        ],
        raw_input_contents=raw_input_contents,
    )

    model_infer_method = server.invoke_unary_unary(
        method_descriptor=(
            grpc_predict_v2_pb2.DESCRIPTOR.services_by_name[
                "GRPCInferenceService"
            ].methods_by_name["ModelInfer"]
        ),
        invocation_metadata={},
        request=request,
        timeout=20,
    )

    response, _, code, _ = model_infer_method.termination()
    response = await response
    response_dict = MessageToDict(
        response,
        preserving_proto_field_name=True,
    )
    assert code == grpc.StatusCode.OK
    assert response_dict == {
        "model_name": "TestModel",
        "id": "123",
        "outputs": [
            {
                "name": "fp32_output",
                "datatype": "FP32",
                "shape": ["8"],
                "parameters": {"binary_data_size": {"int64_param": "32"}},
            },
            {
                "name": "int32_output",
                "datatype": "INT32",
                "shape": ["8"],
                "parameters": {"binary_data_size": {"int64_param": "32"}},
            },
            {
                "name": "string_output",
                "datatype": "BYTES",
                "shape": ["8"],
                "parameters": {"binary_data_size": {"int64_param": "59"}},
            },
            {
                "name": "uint8_output",
                "datatype": "UINT8",
                "shape": ["8"],
                "parameters": {"binary_data_size": {"int64_param": "8"}},
            },
            {
                "name": "bool_input",
                "datatype": "BOOL",
                "shape": ["8"],
                "parameters": {"binary_data_size": {"int64_param": "8"}},
            },
        ],
        "raw_output_contents": [
            "mpnZQDMzM0CamZlAMzOzPwAAwECamVlAAACQQM3MzD8=",
            "BgAAAAIAAAAEAAAAAQAAAAYAAAADAAAABAAAAAEAAAA=",
            "AwAAAENhdAMAAABEb2cEAAAAV29sZgMAAABDYXQDAAAARG9nBAAAAFdvbGYDAAAARG9nBAAAAFdvbGY=",
            "BgIEAQYDBAE=",
            "AQABAAEAAQA=",
        ],
    }
    infer_response = InferResponse.from_grpc(response)
    assert np.array_equal(infer_response.outputs[0].as_numpy(), fp32_data)
    assert np.array_equal(infer_response.outputs[1].as_numpy(), int32_data)
    assert np.array_equal(infer_response.outputs[2].as_numpy(), str_data)
    assert np.array_equal(infer_response.outputs[3].as_numpy(), uint8_data)
    assert np.array_equal(infer_response.outputs[4].as_numpy(), bool_data)


@pytest.mark.asyncio
@patch(
    "kserve.protocol.grpc.servicer.to_headers", return_value=[]
)  # To avoid NotImplementedError from trailing_metadata function
async def test_grpc_fp16_output(mock_to_headers, server):
    """
    If the output contains FP16 datatype, then the outputs should be returned as raw outputs.
    """
    fp32_data = [6.8, 2.8, 4.8, 1.4, 6.0, 3.4, 4.5, 1.6]
    request = grpc_predict_v2_pb2.ModelInferRequest(
        model_name="FP16OutputModel",
        id="123",
        inputs=[
            {
                "name": "fp32_input",
                "shape": [2, 4],
                "datatype": "FP32",
                "contents": {"fp32_contents": fp32_data},
            },
        ],
    )

    model_infer_method = server.invoke_unary_unary(
        method_descriptor=(
            grpc_predict_v2_pb2.DESCRIPTOR.services_by_name[
                "GRPCInferenceService"
            ].methods_by_name["ModelInfer"]
        ),
        invocation_metadata={},
        request=request,
        timeout=20,
    )

    response, _, code, _ = model_infer_method.termination()
    response = await response
    response_dict = MessageToDict(
        response,
        preserving_proto_field_name=True,
    )
    assert code == grpc.StatusCode.OK
    assert response_dict == {
        "model_name": "FP16OutputModel",
        "id": "123",
        "outputs": [
            {
                "name": "fp16_output",
                "datatype": "FP16",
                "shape": ["8"],
                "parameters": {"binary_data_size": {"int64_param": "16"}},
            },
            {
                "name": "fp32_output",
                "datatype": "FP32",
                "shape": ["8"],
                "parameters": {"binary_data_size": {"int64_param": "32"}},
            },
        ],
        "raw_output_contents": [
            "zUaaQc1Emj0ARs1CgERmPg==",
            "mpnZQDMzM0CamZlAMzOzPwAAwECamVlAAACQQM3MzD8=",
        ],
    }
    infer_response = InferResponse.from_grpc(response)
    assert np.array_equal(
        infer_response.outputs[0].as_numpy(), np.array(fp32_data, dtype=np.float16)
    )
    assert np.array_equal(
        infer_response.outputs[1].as_numpy(), np.array(fp32_data, dtype=np.float32)
    )


@pytest.mark.asyncio
@patch(
    "kserve.protocol.grpc.servicer.to_headers", return_value=[]
)  # To avoid NotImplementedError from trailing_metadata function
async def test_grpc_fp16_input(mock_to_headers, server):
    fp16_data = np.array([6.8, 2.8, 4.8, 1.4, 6.0, 3.4, 4.5, 1.6], dtype=np.float16)
    raw_input_contents = [fp16_data.tobytes()]
    request = grpc_predict_v2_pb2.ModelInferRequest(
        model_name="FP16InputModel",
        id="123",
        inputs=[
            {
                "name": "fp16_input",
                "shape": [2, 4],
                "datatype": "FP16",
            },
        ],
        raw_input_contents=raw_input_contents,
    )

    model_infer_method = server.invoke_unary_unary(
        method_descriptor=(
            grpc_predict_v2_pb2.DESCRIPTOR.services_by_name[
                "GRPCInferenceService"
            ].methods_by_name["ModelInfer"]
        ),
        invocation_metadata={},
        request=request,
        timeout=20,
    )

    response, _, code, _ = model_infer_method.termination()
    response = await response
    response_dict = MessageToDict(
        response,
        preserving_proto_field_name=True,
    )
    assert code == grpc.StatusCode.OK
    assert response_dict == {
        "model_name": "FP16InputModel",
        "id": "123",
        "outputs": [
            {
                "name": "int32_output",
                "datatype": "INT64",
                "shape": ["8"],
                "parameters": {"binary_data_size": {"int64_param": "64"}},
            },
            {
                "name": "fp16_output",
                "datatype": "FP16",
                "shape": ["8"],
                "parameters": {"binary_data_size": {"int64_param": "16"}},
            },
        ],
        "raw_output_contents": [
            "AQAAAAAAAAACAAAAAAAAAAMAAAAAAAAABAAAAAAAAAAFAAAAAAAAAAYAAAAAAAAABwAAAAAAAAAIAAAAAAAAAA==",
            "zUaaQc1Emj0ARs1CgERmPg==",
        ],
    }
    infer_response = InferResponse.from_grpc(response)
    assert np.array_equal(infer_response.outputs[1].as_numpy(), fp16_data)


@pytest.mark.asyncio
@patch(
    "kserve.protocol.grpc.servicer.to_headers", return_value=[]
)  # To avoid NotImplementedError from trailing_metadata function
async def test_grpc_raw_inputs_with_missing_input_data(mock_to_headers, server):
    """
    Server should raise InvalidInput if raw_input_contents missing some input data.
    """
    raw_input_contents = [
        np.array([6.8, 2.8, 4.8, 1.4, 6.0, 3.4, 4.5, 1.6], dtype=np.float32).tobytes(),
        np.array([6, 2, 4, 1, 6, 3, 4, 1], dtype=np.int32).tobytes(),
    ]
    request = grpc_predict_v2_pb2.ModelInferRequest(
        model_name="TestModel",
        id="123",
        inputs=[
            {
                "name": "fp32_input",
                "shape": [2, 4],
                "datatype": "FP32",
            },
            {
                "name": "int32_input",
                "shape": [2, 4],
                "datatype": "INT32",
            },
            {
                "name": "string_input",
                "shape": [8],
                "datatype": "BYTES",
            },
        ],
        raw_input_contents=raw_input_contents,
    )

    model_infer_method = server.invoke_unary_unary(
        method_descriptor=(
            grpc_predict_v2_pb2.DESCRIPTOR.services_by_name[
                "GRPCInferenceService"
            ].methods_by_name["ModelInfer"]
        ),
        invocation_metadata={},
        request=request,
        timeout=20,
    )

    with pytest.raises(InvalidInput):
        response, _, _, _ = model_infer_method.termination()
        _ = await response


@pytest.mark.asyncio
@patch(
    "kserve.protocol.grpc.servicer.to_headers", return_value=[]
)  # To avoid NotImplementedError from trailing_metadata function
async def test_grpc_raw_inputs_with_contents_specified(mock_to_headers, server):
    """
    Server should raise InvalidInput if both contents and raw_input_contents specified.
    """
    raw_input_contents = [
        np.array([6.8, 2.8, 4.8, 1.4, 6.0, 3.4, 4.5, 1.6], dtype=np.float32).tobytes(),
        np.array([6, 2, 4, 1, 6, 3, 4, 1], dtype=np.int32).tobytes(),
        serialize_byte_tensor(
            np.array(
                [b"Cat", b"Dog", b"Wolf", b"Cat", b"Dog", b"Wolf", b"Dog", b"Wolf"],
                dtype=np.object_,
            )
        ).item(),
        np.array([6, 2, 4, 1, 6, 3, 4, 1], dtype=np.uint8).tobytes(),
        np.array(
            [True, False, True, False, True, False, True, False], dtype=np.bool_
        ).tobytes(),
    ]
    request = grpc_predict_v2_pb2.ModelInferRequest(
        model_name="TestModel",
        id="123",
        inputs=[
            {
                "name": "fp32_input",
                "shape": [2, 4],
                "datatype": "FP32",
            },
            {
                "name": "int32_input",
                "shape": [2, 4],
                "datatype": "INT32",
            },
            {
                "name": "string_input",
                "shape": [8],
                "datatype": "BYTES",
            },
            {
                "name": "uint8_input",
                "shape": [2, 4],
                "datatype": "UINT8",
                "contents": {
                    "uint_contents": [6, 2, 4, 1, 6, 3, 4, 1],
                },
            },
            {
                "name": "bool_input",
                "shape": [8],
                "datatype": "BOOL",
            },
        ],
        raw_input_contents=raw_input_contents,
    )

    model_infer_method = server.invoke_unary_unary(
        method_descriptor=(
            grpc_predict_v2_pb2.DESCRIPTOR.services_by_name[
                "GRPCInferenceService"
            ].methods_by_name["ModelInfer"]
        ),
        invocation_metadata={},
        request=request,
        timeout=20,
    )

    with pytest.raises(InvalidInput):
        response, _, _, _ = model_infer_method.termination()
        _ = await response



# Copyright 2025 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Tuple

import pytest

from kserve.protocol.rest.openai import (
    OpenAIEncoderModel,
)
from kserve.protocol.rest.openai.types import (
    Embedding,
    EmbeddingCompletionRequest as EmbeddingRequest,
    RerankRequest,
    Rerank,
)

FIXTURES_PATH = Path(__file__).parent / "fixtures" / "openai"


class DummyModel(OpenAIEncoderModel):
    embedding_data: Tuple[Embedding]
    rerank_data: Tuple[Rerank]

    def __init__(self, embedding_data: Tuple[Embedding], rerank_data: Tuple[Rerank]):
        super().__init__("dummy-model")
        self.embedding_data = embedding_data
        self.rerank_data = rerank_data

    async def create_embedding(self, request: EmbeddingRequest) -> Embedding:
        return self.embedding_data[0]

    async def create_rerank(self, request: RerankRequest) -> Rerank:
        return self.rerank_data[0]


@pytest.fixture
def embedding():
    with open(FIXTURES_PATH / "embedding.json") as f:
        return Embedding.model_validate_json(f.read())


@pytest.fixture
def embedding_create_params():
    with open(FIXTURES_PATH / "embedding_create_params.json") as f:
        return EmbeddingRequest.model_validate_json(f.read())


@pytest.fixture
def rerank():
    with open(FIXTURES_PATH / "rerank.json") as f:
        return Rerank.model_validate_json(f.read())


@pytest.fixture
def rerank_create_params():
    with open(FIXTURES_PATH / "rerank_create_params.json") as f:
        return RerankRequest.model_validate_json(f.read())


@pytest.fixture
def dummy_model_encoder(embedding, rerank):
    return DummyModel(embedding_data=(embedding,), rerank_data=(rerank,))


class TestOpenAICreateEmbedding:
    @pytest.mark.asyncio
    async def test_create_embedding(
        self,
        dummy_model_encoder: DummyModel,
        embedding: Embedding,
        embedding_create_params: EmbeddingRequest,
    ):
        c = await dummy_model_encoder.create_embedding(embedding_create_params)
        assert isinstance(c, Embedding)
        assert c.model_dump_json(indent=2) == embedding.model_dump_json(indent=2)

    @pytest.mark.asyncio
    async def test_create_rerank(
        self,
        dummy_model_encoder: DummyModel,
        rerank: Rerank,
        rerank_create_params: RerankRequest,
    ):
        c = await dummy_model_encoder.create_rerank(rerank_create_params)
        assert isinstance(c, Rerank)
        assert c.model_dump_json(indent=2) == rerank.model_dump_json(indent=2)



# Copyright 2024 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Global context management for KServe model server configuration."""

import contextvars
from typing import Optional

from .predictor_config import PredictorConfig

# Global context variable for predictor configuration
_predictor_config_var: contextvars.ContextVar[Optional[PredictorConfig]] = (
    contextvars.ContextVar("predictor_config", default=None)
)


def set_predictor_config(config: PredictorConfig) -> None:
    """Set the predictor configuration for the current context.

    This should be called once during ModelServer initialization.

    Args:
        config: The PredictorConfig instance to set
    """
    _predictor_config_var.set(config)


def get_predictor_config() -> Optional[PredictorConfig]:
    """Get the predictor configuration from the current context.

    Returns:
        The PredictorConfig instance if set, None otherwise
    """
    return _predictor_config_var.get()



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import datetime
import io
import json
import os
import re
from typing import Dict
from unittest import mock

import avro.io
import avro.schema
import httpx
import numpy as np
import pandas as pd
import pytest
import pytest_asyncio
from cloudevents.conversion import to_binary, to_structured
from cloudevents.http import CloudEvent
from fastapi.testclient import TestClient
from ray import serve

from kserve import Model, ModelRepository, ModelServer
from kserve.constants.constants import (
    FASTAPI_APP_IMPORT_STRING,
    INFERENCE_CONTENT_LENGTH_HEADER,
)
from kserve.errors import InvalidInput, NoModelReady
from kserve.model import PredictorProtocol
from kserve.model_server import app as kserve_app
from kserve.ray import RayModel
from kserve.protocol.infer_type import (
    InferInput,
    InferOutput,
    InferRequest,
    InferResponse,
    RequestedOutput,
)
from kserve.utils.utils import generate_uuid, get_predict_input, get_predict_response
from kserve.protocol.dataplane import DataPlane
from kserve.protocol.model_repository_extension import ModelRepositoryExtension
from kserve.protocol.rest.multiprocess.server import RESTServerMultiProcess
from kserve.predictor_config import PredictorConfig
from kserve import context as kserve_context

test_avsc_schema = """
        {
        "namespace": "example.avro",
         "type": "record",
         "name": "User",
         "fields": [
             {"name": "name", "type": "string"},
             {"name": "favorite_number",  "type": ["int", "null"]},
             {"name": "favorite_color", "type": ["string", "null"]}
         ]
        }
        """

fake_stream_data = "some streamed data"


def dummy_cloud_event(
    data,
    set_contenttype: bool = False,
    add_extension: bool = False,
    contenttype: str = "application/json",
):
    # This data defines a binary cloudevent
    attributes = {
        "type": "com.example.sampletype1",
        "source": "https://example.com/event-producer",
        "specversion": "1.0",
        "id": "36077800-0c23-4f38-a0b4-01f4369f670a",
        "time": "2021-01-28T21:04:43.144141+00:00",
    }
    if set_contenttype:
        attributes["content-type"] = contenttype
    if add_extension:
        attributes["custom-extension"] = "custom-value"

    event = CloudEvent(attributes, data)
    return event


async def fake_data_streamer():
    for _ in range(10):
        yield fake_stream_data.encode()
        await asyncio.sleep(0.5)  # sleep 1/2 second


class DummyStreamModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        return fake_data_streamer()


class TestStreamPredict:

    @pytest_asyncio.fixture(scope="class")
    async def app(self, server):  # pylint: disable=no-self-use
        model = DummyStreamModel("TestModel")
        model.load()
        server.register_model(model)
        yield kserve_app
        await server.model_repository_extension.unload("TestModel")

    @pytest.fixture(scope="class")
    def http_server_client(self, app):
        return TestClient(app)

    def test_predict_stream(self, http_server_client):
        with http_server_client.stream(
            "POST", "/v1/models/TestModel:predict", content=b'{"instances":[[1,2]]}'
        ) as response:
            response: httpx.Response
            all_data = []
            for value in response.iter_bytes():
                data = value.decode()
                assert fake_stream_data in data
                all_data.append(data)
        assert all(
            [fake_stream_data in data for data in all_data]
        ), "Unexpected number of streamed responses"


class DummyModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        if isinstance(request, InferRequest):
            inputs = get_predict_input(request)
            infer_response = get_predict_response(request, inputs, self.name)
            if request.parameters:
                infer_response.parameters = request.parameters
            if request.inputs[0].parameters:
                infer_response.outputs[0].parameters = request.inputs[0].parameters
            return infer_response
        else:
            if "inputs" in request:
                return {"predictions": request["inputs"]}
            else:
                return {"predictions": request["instances"]}

    async def explain(self, request, headers=None):
        if "inputs" in request:
            return {"predictions": request["inputs"]}
        else:
            return {"predictions": request["instances"]}


@serve.deployment
class DummyServeModel(Model):
    async def predict(self, request, headers=None):
        if isinstance(request, InferRequest):
            inputs = get_predict_input(request)
            infer_response = get_predict_response(request, inputs, self.name)
            return infer_response
        else:
            return {"predictions": request["instances"]}

    async def explain(self, request, headers=None):
        return {"predictions": request["instances"]}


class DummyCEModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        return {"predictions": request["instances"]}

    async def explain(self, request, headers=None):
        return {"predictions": request["instances"]}


class DummyAvroCEModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    def _parserequest(self, request):
        schema = avro.schema.parse(test_avsc_schema)
        raw_bytes = request
        bytes_reader = io.BytesIO(raw_bytes)
        decoder = avro.io.BinaryDecoder(bytes_reader)
        reader = avro.io.DatumReader(schema)
        record1 = reader.read(decoder)
        return record1

    def preprocess(self, request, headers: Dict[str, str] = None):
        assert headers["ce-specversion"] == "1.0"
        assert headers["ce-source"] == "https://example.com/event-producer"
        assert headers["ce-type"] == "com.example.sampletype1"
        assert headers["ce-content-type"] == "application/avro"
        return self._parserequest(request)

    async def predict(self, request, headers=None):
        return {
            "predictions": [
                [request["name"], request["favorite_number"], request["favorite_color"]]
            ]
        }

    async def explain(self, request, headers=None):
        return {
            "predictions": [
                [request["name"], request["favorite_number"], request["favorite_color"]]
            ]
        }


class DummyModelRepository(ModelRepository):
    def __init__(self, test_load_success: bool, fail_with_exception: bool = False):
        super().__init__()
        self.test_load_success = test_load_success
        self.fail_with_exception = fail_with_exception

    async def load(self, name: str) -> bool:
        if self.test_load_success:
            model = DummyModel(name)
            model.load()
            self.update(model)
            return model.ready
        else:
            if self.fail_with_exception:
                raise Exception(f"Could not load model {name}.")
            else:
                return False


class DummyNeverReadyModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False


class DummyFP16OutputModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        outputs = pd.DataFrame(
            {
                "fp16_output": request.get_input_by_name("fp32_input")
                .as_numpy()
                .astype(np.float16)
                .flatten(),
                "fp32_output": request.get_input_by_name("fp32_input")
                .as_numpy()
                .flatten(),
            }
        )
        # Fixme: Gets only the 1st element of the input
        # inputs = get_predict_input(request)
        infer_response = get_predict_response(request, outputs, self.name)
        if request.parameters:
            infer_response.parameters = request.parameters
            infer_response.parameters.pop("binary_data_output", None)
        if request.inputs[0].parameters:
            infer_response.outputs[0].parameters = request.inputs[0].parameters
            infer_response.outputs[0].parameters.pop("binary_data", None)
        return infer_response


class DummyFP16InputModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        outputs = pd.DataFrame(
            {
                "str_output": request.get_input_by_name("str_input")
                .as_numpy()
                .flatten(),
                "fp32_output": request.get_input_by_name("fp16_input")
                .as_numpy()
                .astype(np.float32)
                .flatten(),
            }
        )
        # Fixme: Gets only the 1st element of the input
        # inputs = get_predict_input(request)
        infer_response = get_predict_response(request, outputs, self.name)
        if request.parameters:
            infer_response.parameters = request.parameters
            infer_response.parameters.pop("binary_data_output", None)
        if request.inputs[0].parameters:
            infer_response.outputs[0].parameters = request.inputs[0].parameters
            infer_response.outputs[0].parameters.pop("binary_data", None)
        return infer_response


class DateTimeModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, payload, headers=None, response_headers=None):
        if isinstance(payload, InferRequest):
            response_id = generate_uuid()
            infer_output = InferOutput(
                name="output-0",
                shape=[1],
                datatype="BYTES",
            )
            infer_output.set_data_from_numpy(
                np.array(
                    [datetime.datetime.now(tz=datetime.timezone.utc)], dtype=np.object_
                ),
                binary_data=False,
            )
            return InferResponse(
                model_name=self.name,
                infer_outputs=[infer_output],
                response_id=response_id,
            )
        return {"predictions": [datetime.datetime.now(tz=datetime.timezone.utc)]}

    async def explain(self, payload, headers=None):
        return {"predictions": [datetime.datetime.now(tz=datetime.timezone.utc)]}


@pytest.fixture(scope="module")
def http_server_client():
    return TestClient(kserve_app, headers={"content-type": "application/json"})


@pytest.mark.asyncio
class TestModel:
    async def test_validate(self):
        model = DummyModel("TestModel")

        # Test REST_V1 protocol validation
        predictor_config_v1 = PredictorConfig(
            predictor_host="localhost",
            predictor_protocol=PredictorProtocol.REST_V1.value,
        )
        kserve_context.set_predictor_config(predictor_config_v1)

        good_request = {"instances": []}
        validated_request = model.validate(good_request)
        assert validated_request == good_request
        bad_request = {"instances": "invalid"}
        with pytest.raises(InvalidInput):
            model.validate(bad_request)

        # Test REST_V2 protocol validation
        predictor_config_v2 = PredictorConfig(
            predictor_host="localhost",
            predictor_protocol=PredictorProtocol.REST_V2.value,
        )
        kserve_context.set_predictor_config(predictor_config_v2)

        good_request = {"inputs": []}
        validated_request = model.validate(good_request)
        assert validated_request == good_request
        bad_request = {"inputs": "invalid"}
        with pytest.raises(InvalidInput):
            model.validate(bad_request)


# Separate out v1 and v2 endpoint unit tests in
# https://github.com/kserve/kserve/blob/master/python/kserve/test/test_server.py.


class TestV1Endpoints:

    @pytest_asyncio.fixture(scope="class", autouse=True)
    async def app(self, server):
        model = DummyModel("TestModel")
        model.load()
        server.register_model(model)
        datetime_model = DateTimeModel("DateTimeModel")
        datetime_model.load()
        server.register_model(datetime_model)
        yield kserve_app
        await server.model_repository_extension.unload("TestModel")
        await server.model_repository_extension.unload("DateTimeModel")

    def test_liveness_v1(self, http_server_client):
        resp = http_server_client.get("/")
        assert resp.status_code == 200
        assert resp.json() == {"status": "alive"}

    def test_model_v1(self, http_server_client):
        resp = http_server_client.get("/v1/models/TestModel")
        assert resp.status_code == 200

    def test_unknown_model_v1(self, http_server_client):
        resp = http_server_client.get("/v1/models/InvalidModel")
        assert resp.status_code == 404
        assert resp.json() == {"error": "Model with name InvalidModel does not exist."}

    def test_list_models_v1(self, http_server_client):
        resp = http_server_client.get("/v1/models")
        assert resp.status_code == 200
        assert resp.json() == {"models": ["TestModel", "DateTimeModel"]}

    def test_predict_v1(self, http_server_client):
        resp = http_server_client.post(
            "/v1/models/TestModel:predict", content=b'{"instances":[[1,2]]}'
        )
        assert resp.status_code == 200
        assert resp.content == b'{"predictions":[[1,2]]}'
        assert resp.headers["content-type"] == "application/json"

    def test_explain_v1(self, http_server_client):
        resp = http_server_client.post(
            "/v1/models/TestModel:explain", content=b'{"instances":[[1,2]]}'
        )
        assert resp.status_code == 200
        assert resp.content == b'{"predictions":[[1,2]]}'
        assert resp.headers["content-type"] == "application/json"

    def test_unknown_path_v1(self, http_server_client):
        resp = http_server_client.get("/unknown_path")
        assert resp.status_code == 404
        assert resp.json() == {"detail": "Not Found"}

    def test_metrics_v1(self, http_server_client):
        resp = http_server_client.get("/metrics")
        assert resp.status_code == 200
        assert resp.content is not None

    def test_datetime_output_v1(self, http_server_client):
        resp = http_server_client.post(
            "/v1/models/DateTimeModel:predict", content=b'{"instances":[[1,2]]}'
        )
        assert resp.status_code == 200
        assert resp.headers["content-type"] == "application/json"
        result = json.loads(resp.content)
        assert isinstance(result["predictions"][0], str)
        result_datetime = datetime.datetime.fromisoformat(result["predictions"][0])
        datetime_now = datetime.datetime.now(tz=datetime.timezone.utc)
        assert result_datetime.date() == datetime_now.date()
        assert result_datetime.tzinfo == datetime_now.tzinfo


class TestV2Endpoints:

    @pytest_asyncio.fixture(scope="class", autouse=True)
    async def app(self, server):
        model = DummyModel("TestModel")
        model.load()
        server.register_model(model)
        fp16_input_model = DummyFP16InputModel("FP16InputModel")
        fp16_input_model.load()
        server.register_model(fp16_input_model)
        fp16_output_model = DummyFP16OutputModel("FP16OutputModel")
        fp16_output_model.load()
        server.register_model(fp16_output_model)
        datetime_model = DateTimeModel("DateTimeModel")
        datetime_model.load()
        server.register_model(datetime_model)
        yield kserve_app
        await server.model_repository_extension.unload("TestModel")
        await server.model_repository_extension.unload("FP16InputModel")
        await server.model_repository_extension.unload("FP16OutputModel")
        await server.model_repository_extension.unload("DateTimeModel")

    def test_list_models_v2(self, http_server_client):
        resp = http_server_client.get("/v2/models")
        assert resp.status_code == 200
        assert resp.json() == {
            "models": [
                "TestModel",
                "FP16InputModel",
                "FP16OutputModel",
                "DateTimeModel",
            ]
        }

    def test_infer_v2(self, http_server_client):
        input_data = b'{"inputs": [{"name": "input-0","shape": [1, 2],"datatype": "INT32","data": [[1,2]]}]}'
        resp = http_server_client.post(
            "/v2/models/TestModel/infer",
            content=input_data,
            headers={"content-type": "application/json"},
        )

        result = json.loads(resp.content)
        assert resp.status_code == 200
        assert result["outputs"][0]["data"] == [1, 2]
        assert resp.headers["content-type"] == "application/json"

    def test_infer_parameters_v2(self, http_server_client):
        model_name = "TestModel"
        req = InferRequest(
            model_name=model_name,
            request_id="123",
            parameters={
                "test-str": "dummy",
                "test-bool": True,
                "test-int": 100,
                "test-float": 1.3,
            },
            infer_inputs=[
                InferInput(
                    name="input-0",
                    datatype="INT32",
                    shape=[1, 2],
                    data=[1, 2],
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                        "test-float": 1.3,
                    },
                )
            ],
        )
        infer_dict, _ = req.to_rest()
        input_data = json.dumps(infer_dict).encode("utf-8")
        expected_res = InferResponse(
            model_name=model_name,
            response_id="123",
            parameters={
                "test-str": "dummy",
                "test-bool": True,
                "test-int": 100,
                "test-float": 1.3,
            },
            infer_outputs=[
                InferOutput(
                    name="output-0",
                    datatype="INT32",
                    shape=[1, 2],
                    data=[1, 2],
                    parameters={
                        "test-str": "dummy",
                        "test-bool": True,
                        "test-int": 100,
                        "test-float": 1.3,
                    },
                )
            ],
        )
        resp = http_server_client.post("/v2/models/TestModel/infer", content=input_data)
        assert resp.status_code == 200
        assert resp.headers["content-type"] == "application/json"
        result = InferResponse.from_rest(response=json.loads(resp.content))
        assert result == expected_res

    def test_fp16_input_as_binary_data(self, http_server_client):
        fp16_data = np.array(
            [[6.8, 2.8, 4.8, 1.4], [6.0, 3.4, 4.5, 1.6]], dtype=np.float16
        )
        str_data = np.array(
            [["cat", "dog", "cat", "dog"], ["cat", "dog", "cat", "dog"]],
            dtype=np.object_,
        )
        fp16_input = InferInput(
            name="fp16_input",
            shape=[2, 4],
            datatype="FP16",
        )
        fp16_input.set_data_from_numpy(fp16_data, binary_data=True)
        request = InferRequest(
            model_name="FP16InputModel",
            request_id="123",
            infer_inputs=[
                fp16_input,
                InferInput(
                    name="str_input",
                    shape=[2, 4],
                    datatype="BYTES",
                    data=str_data.tolist(),
                ),
            ],
        )
        req_bytes, json_length = request.to_rest()
        assert isinstance(req_bytes, bytes)
        resp = http_server_client.post(
            "/v2/models/FP16InputModel/infer",
            content=req_bytes,
            headers={
                INFERENCE_CONTENT_LENGTH_HEADER: str(json_length),
                "Content-Type": "application/octet-stream",
            },
        )
        assert resp.status_code == 200
        assert (
            resp.content
            == b'{"model_name":"FP16InputModel","model_version":null,"id":"123","parameters":null,"outputs":[{"name":"str_output","shape":[8],"datatype":"BYTES","parameters":null,"data":["cat","dog","cat","dog","cat","dog","cat","dog"]},{"name":"fp32_output","shape":[8],"datatype":"FP32","parameters":null,"data":[6.80078125,2.80078125,4.80078125,1.400390625,6.0,3.400390625,4.5,1.599609375]}]}'
        )

    def test_fp16_input_not_binary_data(self, http_server_client):
        fp16_data = np.array(
            [[6.8, 2.8, 4.8, 1.4], [6.0, 3.4, 4.5, 1.6]], dtype=np.float16
        )
        str_data = np.array(
            [["cat", "dog", "cat", "dog"], ["cat", "dog", "cat", "dog"]],
            dtype=np.object_,
        )
        req_dict = {
            "model_name": "FP16InputModel",
            "request_id": "123",
            "inputs": [
                {
                    "name": "fp16_input",
                    "shape": [2, 4],
                    "datatype": "FP16",
                    "data": fp16_data.tolist(),
                },
                {
                    "name": "str_input",
                    "shape": [2, 4],
                    "datatype": "BYTES",
                    "data": str_data.tolist(),
                },
            ],
        }
        resp = http_server_client.post(
            "/v2/models/FP16InputModel/infer",
            json=req_dict,
            headers={
                "Content-Type": "application/json",
            },
        )
        assert resp.status_code == 400

    def test_fp16_output_as_binary_data(self, http_server_client):
        fp32_data = np.array(
            [[6.8, 2.8, 4.8, 1.4], [6.0, 3.4, 4.5, 1.6]], dtype=np.float32
        )
        request = InferRequest(
            model_name="FP16OutputModel",
            request_id="123",
            infer_inputs=[
                InferInput(
                    name="fp32_input",
                    shape=[2, 4],
                    datatype="FP32",
                    data=fp32_data.tolist(),
                )
            ],
            request_outputs=[
                RequestedOutput(
                    name="fp16_output",
                    parameters={"binary_data": True},
                ),
                RequestedOutput(
                    name="fp32_output",
                    parameters={"binary_data": False},
                ),
            ],
        )
        req_dict, _ = request.to_rest()
        resp = http_server_client.post(
            "/v2/models/FP16OutputModel/infer",
            json=req_dict,
            headers={
                "Content-Type": "application/json",
            },
        )
        assert resp.status_code == 200
        assert (
            resp.content
            == b'{"id":"123","model_name":"FP16OutputModel","model_version":null,"outputs":[{"name":"fp16_output","shape":[8],"datatype":"FP16","parameters":{"binary_data_size":16}},{"name":"fp32_output","shape":[8],"datatype":"FP32","data":[6.800000190734863,2.799999952316284,4.800000190734863,1.399999976158142,6.0,3.4000000953674316,4.5,1.600000023841858]}]}\xcdF\x9aA\xcdD\x9a=\x00F\xcdB\x80Df>'
        )
        assert resp.headers.get(INFERENCE_CONTENT_LENGTH_HEADER) == "345"

    def test_fp16_output_not_binary_data(self, http_server_client):
        fp32_data = np.array(
            [[6.8, 2.8, 4.8, 1.4], [6.0, 3.4, 4.5, 1.6]], dtype=np.float32
        )
        req_dict = {
            "model_name": "FP16OutputModel",
            "request_id": "123",
            "inputs": [
                {
                    "name": "fp32_input",
                    "shape": [2, 4],
                    "datatype": "FP32",
                    "data": fp32_data.tolist(),
                }
            ],
            "outputs": [
                {
                    "name": "fp16_output",
                },
                {
                    "name": "fp32_output",
                    "parameters": {"binary_data": False},
                },
            ],
        }
        resp = http_server_client.post(
            "/v2/models/FP16OutputModel/infer",
            json=req_dict,
            headers={
                "Content-Type": "application/json",
            },
        )
        assert resp.status_code == 400

    def test_requested_output(self, http_server_client):
        fp32_data = np.array(
            [[6.8, 2.8, 4.8, 1.4], [6.0, 3.4, 4.5, 1.6]], dtype=np.float32
        )
        request = InferRequest(
            model_name="FP16OutputModel",
            request_id="123",
            infer_inputs=[
                InferInput(
                    name="fp32_input",
                    shape=[2, 4],
                    datatype="FP32",
                    data=fp32_data.tolist(),
                )
            ],
            request_outputs=[
                RequestedOutput(
                    name="fp32_output",
                    parameters={"binary_data": False},
                )
            ],
        )
        req_dict, _ = request.to_rest()
        resp = http_server_client.post(
            "/v2/models/FP16OutputModel/infer",
            json=req_dict,
            headers={
                "Content-Type": "application/json",
            },
        )
        assert resp.status_code == 200
        assert (
            resp.content
            == b'{"model_name":"FP16OutputModel","model_version":null,"id":"123","parameters":null,"outputs":[{"name":"fp32_output","shape":[8],"datatype":"FP32","parameters":null,"data":[6.800000190734863,2.799999952316284,4.800000190734863,1.399999976158142,6.0,3.4000000953674316,4.5,1.600000023841858]}]}'
        )

    def test_all_output_as_binary_data(self, http_server_client):
        fp32_data = np.array(
            [[6.8, 2.8, 4.8, 1.4], [6.0, 3.4, 4.5, 1.6]], dtype=np.float32
        )
        request = InferRequest(
            model_name="FP16OutputModel",
            request_id="123",
            infer_inputs=[
                InferInput(
                    name="fp32_input",
                    shape=[2, 4],
                    datatype="FP32",
                    data=fp32_data.tolist(),
                )
            ],
            parameters={"binary_data_output": True},
        )
        req_dict, _ = request.to_rest()
        resp = http_server_client.post(
            "/v2/models/FP16OutputModel/infer",
            json=req_dict,
            headers={
                "Content-Type": "application/json",
            },
        )
        assert resp.status_code == 200
        assert (
            resp.content
            == b'{"id":"123","model_name":"FP16OutputModel","model_version":null,"outputs":[{"name":"fp16_output","shape":[8],"datatype":"FP16","parameters":{"binary_data_size":16}},{"name":"fp32_output","shape":[8],"datatype":"FP32","parameters":{"binary_data_size":32}}]}\xcdF\x9aA\xcdD\x9a=\x00F\xcdB\x80Df>\x9a\x99\xd9@333@\x9a\x99\x99@33\xb3?\x00\x00\xc0@\x9a\x99Y@\x00\x00\x90@\xcd\xcc\xcc?'
        )
        assert resp.headers.get(INFERENCE_CONTENT_LENGTH_HEADER) == "256"

    def test_binary_data_parameter_precedence(self, http_server_client):
        fp32_data = np.array(
            [[6.8, 2.8, 4.8, 1.4], [6.0, 3.4, 4.5, 1.6]], dtype=np.float32
        )
        request = InferRequest(
            model_name="FP16OutputModel",
            request_id="123",
            infer_inputs=[
                InferInput(
                    name="fp32_input",
                    shape=[2, 4],
                    datatype="FP32",
                    data=fp32_data.tolist(),
                )
            ],
            parameters={"binary_data_output": True},
            request_outputs=[
                RequestedOutput(
                    name="fp16_output",
                    parameters={"binary_data": True},
                ),
                RequestedOutput(
                    name="fp32_output",
                    parameters={"binary_data": False},
                ),
            ],
        )
        req_dict, _ = request.to_rest()
        resp = http_server_client.post(
            "/v2/models/FP16OutputModel/infer",
            json=req_dict,
            headers={
                "Content-Type": "application/json",
            },
        )
        assert resp.status_code == 200
        assert (
            resp.content
            == b'{"id":"123","model_name":"FP16OutputModel","model_version":null,"outputs":[{"name":"fp16_output","shape":[8],"datatype":"FP16","parameters":{"binary_data_size":16}},{"name":"fp32_output","shape":[8],"datatype":"FP32","data":[6.800000190734863,2.799999952316284,4.800000190734863,1.399999976158142,6.0,3.4000000953674316,4.5,1.600000023841858]}]}\xcdF\x9aA\xcdD\x9a=\x00F\xcdB\x80Df>'
        )
        assert resp.headers.get(INFERENCE_CONTENT_LENGTH_HEADER) == "345"

    def test_datetime_output(self, http_server_client):
        input_data = b'{"inputs": [{"name": "input-0","shape": [1, 2],"datatype": "INT32","data": [[1,2]]}]}'
        resp = http_server_client.post(
            "/v2/models/DateTimeModel/infer",
            content=input_data,
        )
        assert resp.status_code == 200
        assert resp.headers["content-type"] == "application/json"
        result = json.loads(resp.content)
        assert isinstance(result["outputs"][0]["data"][0], str)
        result_datetime = datetime.datetime.fromisoformat(
            result["outputs"][0]["data"][0]
        )
        datetime_now = datetime.datetime.now(tz=datetime.timezone.utc)
        assert result_datetime.date() == datetime_now.date()
        assert result_datetime.tzinfo == datetime_now.tzinfo


class TestRayServer:

    @pytest_asyncio.fixture(scope="class", autouse=True)
    async def app(self, server):  # pylint: disable=no-self-use
        serve.start(http_options={"host": "0.0.0.0", "port": 9071})

        # https://github.com/ray-project/ray/blob/releases/2.8.0/python/ray/serve/deployment.py#L256
        model_name = "TestModel"
        ray_app = DummyServeModel.bind(name=model_name)
        handle = serve.run(ray_app, name=model_name, route_prefix="/")

        model = RayModel(model_name, handle=handle)
        model.load()
        server.register_model(model)
        yield kserve_app
        await server.model_repository_extension.unload("TestModel")
        serve.shutdown()

    def test_liveness_handler(self, http_server_client):
        resp = http_server_client.get("/")
        assert resp.status_code == 200
        assert resp.content == b'{"status":"alive"}'

    def test_list_handler(self, http_server_client):
        resp = http_server_client.get("/v1/models")
        assert resp.status_code == 200
        assert resp.content == b'{"models":["TestModel"]}'

    def test_health_handler(self, http_server_client):
        resp = http_server_client.get("/v1/models/TestModel")
        assert resp.status_code == 200
        expected_content = b'{"name":"TestModel","ready":true}'
        assert resp.content == expected_content

    def test_predict(self, http_server_client):
        resp = http_server_client.post(
            "/v1/models/TestModel:predict", content=b'{"instances":[[1,2]]}'
        )
        assert resp.status_code == 200
        assert resp.content == b'{"predictions":[[1,2]]}'
        assert resp.headers["content-type"] == "application/json"

    def test_infer(self, http_server_client):
        input_data = b'{"inputs": [{"name": "input-0","shape": [1, 2],"datatype": "INT32","data": [[1,2]]}]}'
        resp = http_server_client.post("/v2/models/TestModel/infer", content=input_data)

        result = json.loads(resp.content)
        assert resp.status_code == 200
        assert result["outputs"][0]["data"] == [1, 2]
        assert resp.headers["content-type"] == "application/json"

    def test_explain(self, http_server_client):
        resp = http_server_client.post(
            "/v1/models/TestModel:explain", content=b'{"instances":[[1,2]]}'
        )
        assert resp.status_code == 200
        assert resp.content == b'{"predictions":[[1,2]]}'
        assert resp.headers["content-type"] == "application/json"


class TestTFHttpServerModelNotLoaded:

    @pytest_asyncio.fixture(scope="class", autouse=True)
    async def app(self, server):  # pylint: disable=no-self-use
        model = DummyModel("TestModel")
        server.register_model(model)
        yield kserve_app
        await server.model_repository_extension.unload("TestModel")

    def test_model_not_ready_error(self, http_server_client):
        resp = http_server_client.get("/v1/models/TestModel")
        assert resp.status_code == 503


class TestTFHttpServerCloudEvent:

    @pytest_asyncio.fixture(scope="class", autouse=True)
    async def app(self, server):  # pylint: disable=no-self-use
        model = DummyCEModel("TestModel")
        model.load()
        server.register_model(model)
        yield kserve_app
        await server.model_repository_extension.unload("TestModel")

    def test_predict_ce_structured(self, http_server_client):
        event = dummy_cloud_event({"instances": [[1, 2]]})
        headers, body = to_structured(event)

        resp = http_server_client.post(
            "/v1/models/TestModel:predict", headers=headers, content=body
        )
        body = json.loads(resp.content)

        assert resp.status_code == 200
        assert resp.headers["content-type"] == "application/cloudevents+json"

        assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert body["data"] == {"predictions": [[1, 2]]}
        assert body["specversion"] == "1.0"
        assert body["source"] == "io.kserve.inference.TestModel"
        assert body["type"] == "io.kserve.inference.response"
        assert body["time"] > "2021-01-28T21:04:43.144141+00:00"

    def test_predict_custom_ce_attributes(self, http_server_client):
        with mock.patch.dict(
            os.environ,
            {
                "CE_SOURCE": "io.kserve.inference.CustomSource",
                "CE_TYPE": "io.kserve.custom_type",
            },
        ):
            event = dummy_cloud_event({"instances": [[1, 2]]})
            headers, body = to_structured(event)

            resp = http_server_client.post(
                "/v1/models/TestModel:predict", headers=headers, content=body
            )
            body = json.loads(resp.content)

            assert resp.status_code == 200
            assert resp.headers["content-type"] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body["source"] == "io.kserve.inference.CustomSource"
            assert body["type"] == "io.kserve.custom_type"

    def test_predict_merge_structured_ce_attributes(self, http_server_client):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event({"instances": [[1, 2]]}, add_extension=True)
            headers, body = to_structured(event)

            resp = http_server_client.post(
                "/v1/models/TestModel:predict", headers=headers, content=body
            )
            body = json.loads(resp.content)

            assert resp.status_code == 200
            assert resp.headers["content-type"] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body["source"] == "io.kserve.inference.TestModel"
            assert body["type"] == "io.kserve.inference.response"
            assert (
                body["custom-extension"] == "custom-value"
            )  # Added by add_extension=True in dummy_cloud_event
            assert body["time"] > "2021-01-28T21:04:43.144141+00:00"

    def test_predict_merge_binary_ce_attributes(self, http_server_client):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event(
                {"instances": [[1, 2]]}, set_contenttype=True, add_extension=True
            )
            headers, body = to_binary(event)

            resp = http_server_client.post(
                "/v1/models/TestModel:predict", headers=headers, content=body
            )

            assert resp.status_code == 200
            assert resp.headers["content-type"] == "application/json"
            assert resp.headers["ce-specversion"] == "1.0"
            assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            # Added by add_extension=True in dummy_cloud_event
            assert resp.headers["ce-custom-extension"] == "custom-value"
            assert resp.headers["ce-source"] == "io.kserve.inference.TestModel"
            assert resp.headers["ce-type"] == "io.kserve.inference.response"
            assert resp.headers["ce-time"] > "2021-01-28T21:04:43.144141+00:00"
            assert resp.content == b'{"predictions": [[1, 2]]}'

    def test_predict_ce_binary_dict(self, http_server_client):
        event = dummy_cloud_event({"instances": [[1, 2]]}, set_contenttype=True)
        headers, body = to_binary(event)

        resp = http_server_client.post(
            "/v1/models/TestModel:predict", headers=headers, content=body
        )

        assert resp.status_code == 200
        assert resp.headers["content-type"] == "application/json"
        assert resp.headers["ce-specversion"] == "1.0"
        assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert resp.headers["ce-source"] == "io.kserve.inference.TestModel"
        assert resp.headers["ce-type"] == "io.kserve.inference.response"
        assert resp.headers["ce-time"] > "2021-01-28T21:04:43.144141+00:00"
        assert resp.content == b'{"predictions": [[1, 2]]}'

    def test_predict_ce_binary_bytes(self, http_server_client):
        event = dummy_cloud_event(b'{"instances":[[1,2]]}', set_contenttype=True)
        headers, body = to_binary(event)
        resp = http_server_client.post(
            "/v1/models/TestModel:predict", headers=headers, content=body
        )

        assert resp.status_code == 200
        assert resp.headers["content-type"] == "application/json"
        assert resp.headers["ce-specversion"] == "1.0"
        assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert resp.headers["ce-source"] == "io.kserve.inference.TestModel"
        assert resp.headers["ce-type"] == "io.kserve.inference.response"
        assert resp.headers["ce-time"] > "2021-01-28T21:04:43.144141+00:00"
        assert resp.content == b'{"predictions": [[1, 2]]}'

    def test_predict_ce_bytes_bad_format_exception(self, http_server_client):
        event = dummy_cloud_event(b"{", set_contenttype=True)
        headers, body = to_binary(event)

        resp = http_server_client.post(
            "/v1/models/TestModel:predict", headers=headers, content=body
        )

        assert resp.status_code == 400
        error_regex = re.compile(
            "Failed to decode or parse binary json cloudevent: "
            "unexpected end of data:*"
        )
        response = json.loads(resp.content)
        assert error_regex.match(response["error"]) is not None

    def test_predict_ce_bytes_bad_hex_format_exception(self, http_server_client):
        event = dummy_cloud_event(b"0\x80\x80\x06World!\x00\x00", set_contenttype=True)
        headers, body = to_binary(event)

        resp = http_server_client.post(
            "/v1/models/TestModel:predict", headers=headers, content=body
        )

        assert resp.status_code == 400
        error_regex = re.compile(
            "Failed to decode or parse binary json cloudevent: "
            "'utf-8' codec can't decode byte 0x80 in position 1: invalid start byte.*"
        )
        response = json.loads(resp.content)
        assert error_regex.match(response["error"]) is not None


class TestTFHttpServerAvroCloudEvent:

    @pytest_asyncio.fixture(scope="class", autouse=True)
    async def app(self, server):  # pylint: disable=no-self-use
        model = DummyAvroCEModel("TestModel")
        model.load()
        server.register_model(model)
        yield kserve_app
        await server.model_repository_extension.unload("TestModel")

    def test_predict_ce_avro_binary(self, http_server_client):
        schema = avro.schema.parse(test_avsc_schema)
        msg = {"name": "foo", "favorite_number": 1, "favorite_color": "pink"}

        writer = avro.io.DatumWriter(schema)
        bytes_writer = io.BytesIO()
        encoder = avro.io.BinaryEncoder(bytes_writer)
        writer.write(msg, encoder)
        data = bytes_writer.getvalue()

        event = dummy_cloud_event(
            data, set_contenttype=True, contenttype="application/avro"
        )
        # Creates the HTTP request representation of the CloudEvent in binary content mode
        headers, body = to_binary(event)
        resp = http_server_client.post(
            "/v1/models/TestModel:predict", headers=headers, content=body
        )

        assert resp.status_code == 200
        assert resp.headers["content-type"] == "application/json"
        assert resp.headers["ce-specversion"] == "1.0"
        assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert resp.headers["ce-source"] == "io.kserve.inference.TestModel"
        assert resp.headers["ce-type"] == "io.kserve.inference.response"
        assert resp.headers["ce-time"] > "2021-01-28T21:04:43.144141+00:00"
        assert resp.content == b'{"predictions": [["foo", 1, "pink"]]}'


class TestTFHttpServerLoadAndUnLoad:
    @pytest_asyncio.fixture(scope="class", autouse=True)
    def app(self, server):
        mp = pytest.MonkeyPatch()
        mp.setattr(
            server.model_repository_extension,
            "_model_registry",
            DummyModelRepository(test_load_success=True),
        )
        yield
        mp.undo()

    def test_load(self, http_server_client):
        resp = http_server_client.post("/v2/repository/models/model/load", content=b"")
        assert resp.status_code == 200
        assert resp.content == b'{"name":"model","load":true}'

    def test_unload(self, http_server_client):
        resp = http_server_client.post(
            "/v2/repository/models/model/unload", content=b""
        )
        assert resp.status_code == 200
        assert resp.content == b'{"name":"model","unload":true}'


class TestTFHttpServerLoadAndUnLoadFailure:
    @pytest.fixture(scope="class", autouse=True)
    def app(self, server):
        mp = pytest.MonkeyPatch()
        mp.setattr(
            server.model_repository_extension,
            "_model_registry",
            DummyModelRepository(test_load_success=False),
        )
        yield
        mp.undo()

    def test_load_fail(self, http_server_client):
        resp = http_server_client.post("/v2/repository/models/model/load", content=b"")
        assert resp.status_code == 503

    def test_unload_fail(self, http_server_client):
        resp = http_server_client.post(
            "/v2/repository/models/model/unload", content=b""
        )
        assert resp.status_code == 404


class TestTFHttpServerModelNotReady:

    @pytest_asyncio.fixture(scope="class", autouse=True)
    async def app(self, server):  # pylint: disable=no-self-use
        model = DummyModel("TestModel")
        server.register_model(model)
        yield kserve_app
        await server.model_repository_extension.unload("TestModel")

    def test_model_not_ready_v1(self, http_server_client):
        resp = http_server_client.get("/v1/models/TestModel")
        assert resp.status_code == 503

    def test_model_not_ready_v2(self, http_server_client):
        resp = http_server_client.get("/v2/models/TestModel/ready")
        assert resp.status_code == 503

    def test_predict(self, http_server_client):
        resp = http_server_client.post(
            "/v1/models/TestModel:predict", content=b'{"instances":[[1,2]]}'
        )
        assert resp.status_code == 503

    def test_infer(self, http_server_client):
        input_data = b'{"inputs": [{"name": "input-0","shape": [1, 2],"datatype": "INT32","data": [[1,2]]}]}'
        resp = http_server_client.post("/v2/models/TestModel/infer", content=input_data)
        assert resp.status_code == 503

    def test_explain(self, http_server_client):
        resp = http_server_client.post(
            "/v1/models/TestModel:explain", content=b'{"instances":[[1,2]]}'
        )
        assert resp.status_code == 503


class TestWithUnhealthyModel:
    def test_with_not_ready_model(self):
        model = DummyNeverReadyModel("Dummy")
        server = ModelServer()
        with pytest.raises(NoModelReady) as exc_info:
            server.start([model])
        assert exc_info.type == NoModelReady


class TestMutiProcessServer:

    @pytest.mark.asyncio
    async def test_rest_server_multiprocess(self):
        model_repository = ModelRepository()
        dummy_model = DummyModel("TestModel")
        dummy_model.load()
        model_repository.update(dummy_model)
        model_repository.load("TestModel")
        data_plane = DataPlane(model_registry=model_repository)
        model_repository_extension = ModelRepositoryExtension(
            model_registry=model_repository
        )
        http_port = 8080
        workers = 4

        server = RESTServerMultiProcess(
            FASTAPI_APP_IMPORT_STRING,
            data_plane,
            model_repository_extension,
            http_port,
            workers=workers,
        )

        # Start server
        task = asyncio.create_task(server.start())

        # Wait for processes to start and become ready
        max_attempts = 30
        for attempt in range(max_attempts):
            await asyncio.sleep(1)
            if len(server._processes) == workers and all(
                p.is_alive() for p in server._processes
            ):
                break
            if attempt == max_attempts - 1:
                raise RuntimeError("Server worker processes did not start properly")

        try:
            # Constantly poll to check if the port is open
            for _ in range(30):
                try:
                    resp = httpx.get("http://localhost:8080/")
                    if resp.status_code == 200:
                        break
                except httpx.RequestError:
                    await asyncio.sleep(1)
            else:
                raise RuntimeError("Server did not start in time")

            # Send predict request
            input_data = b'{"inputs": [{"name": "input-0","shape": [1, 2],"datatype": "INT32","data": [[1,2]]}]}'
            resp = httpx.post(
                "http://localhost:8080/v2/models/TestModel/infer",
                content=input_data,
                headers={"content-type": "application/json"},
            )
            result = json.loads(resp.content)
            assert resp.status_code == 200
            assert result["outputs"][0]["data"] == [1, 2]
            assert resp.headers["content-type"] == "application/json"

            # Kill a process and check if it's restarted
            old_pid = server._processes[0].pid
            server._processes[0].kill()
            await server._processes[0].wait_for_termination()

            # Wait for respawn
            await asyncio.sleep(5)

            new_pid = server._processes[0].pid
            assert old_pid != new_pid, "Process was not restarted"
        finally:
            # Cleanup
            await server.stop()
            await asyncio.sleep(5)
            task.cancel()


@pytest.mark.asyncio
class TestModelServerWithPredictorConfig:
    """Test ModelServer initialization and functionality with PredictorConfig."""

    async def test_model_server_with_predictor_config(self):
        """Test that ModelServer properly sets PredictorConfig in the global context."""
        # Create a PredictorConfig
        predictor_config = PredictorConfig(
            predictor_host="test-predictor.example.com",
            predictor_protocol=PredictorProtocol.REST_V1.value,
            predictor_request_retries=3,
            predictor_request_timeout_seconds=10,
            predictor_health_check=True,
            predictor_use_ssl=False,
        )

        # Create ModelServer with PredictorConfig
        model_repository = ModelRepository()
        model_server = ModelServer(
            registered_models=model_repository, predictor_config=predictor_config
        )

        # Verify that the predictor config was set in the global context
        context_config = kserve_context.get_predictor_config()
        assert context_config is not None
        assert context_config.predictor_host == "test-predictor.example.com"
        assert context_config.predictor_protocol == PredictorProtocol.REST_V1.value
        assert context_config.predictor_request_retries == 3
        assert context_config.predictor_request_timeout_seconds == 10
        assert context_config.predictor_health_check is True
        assert context_config.predictor_use_ssl is False

        # Verify that the DataPlane can access the config from context
        dataplane = model_server.dataplane
        assert dataplane is not None

        # Test with None predictor_config (should not fail)
        model_server_no_config = ModelServer(
            registered_models=ModelRepository(), predictor_config=None
        )
        assert model_server_no_config.dataplane is not None

    async def test_model_server_dataplane_with_predictor_config(self):
        """Test that the DataPlane created by ModelServer uses the PredictorConfig from context."""
        # Create a PredictorConfig with specific settings
        predictor_config = PredictorConfig(
            predictor_host="grpc-predictor.example.com",
            predictor_protocol=PredictorProtocol.GRPC_V2.value,
            predictor_request_retries=5,
            predictor_request_timeout_seconds=30,
            predictor_health_check=False,  # Disabled to avoid network calls in test
            predictor_use_ssl=True,
        )

        # Create ModelServer with the config
        model_repository = ModelRepository()
        dummy_model = DummyModel("TestModel")
        dummy_model.load()
        model_repository.update(dummy_model)

        model_server = ModelServer(
            registered_models=model_repository, predictor_config=predictor_config
        )

        # Verify the DataPlane was created and can access the config
        dataplane = model_server.dataplane
        assert dataplane is not None

        # Verify the context has the correct config
        context_config = kserve_context.get_predictor_config()
        assert context_config.predictor_host == "grpc-predictor.example.com"
        assert context_config.predictor_protocol == PredictorProtocol.GRPC_V2.value
        assert context_config.predictor_use_ssl is True

        # Test that model operations work
        assert await dataplane.model_ready("TestModel") is True

        # Test inference (should work even with predictor config since health_check is disabled)
        body = b'{"instances":[[1,2]]}'
        infer_request, req_attributes = dataplane.decode(body, None)
        resp, headers = await dataplane.infer("TestModel", infer_request)
        assert resp is not None

    def test_model_server_backwards_compatibility(self):
        """Test that ModelServer works without PredictorConfig for backwards compatibility."""
        # Create ModelServer without predictor_config (should work)
        model_repository = ModelRepository()
        model_server = ModelServer(registered_models=model_repository)

        # Should create DataPlane successfully
        assert model_server.dataplane is not None

        # Context should handle None predictor config gracefully
        # The DataPlane should still be functional even without predictor config



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
KServe

Python SDK for KServe  # noqa: E501

The version of the OpenAPI document: v0.1
Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1beta1_pod_spec import V1beta1PodSpec  # noqa: E501
from kserve.rest import ApiException


class TestV1beta1PodSpec(unittest.TestCase):
    """V1beta1PodSpec unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1beta1PodSpec
        include_option is a boolean, when False only required
        params are included, when True both required and
        optional params are included"""
        # model = kserve.models.v1beta1_pod_spec.V1beta1PodSpec()  # noqa: E501
        if include_optional:
            return V1beta1PodSpec(
                active_deadline_seconds=56,
                affinity=None,
                automount_service_account_token=True,
                containers=[None],
                dns_config=None,
                dns_policy="0",
                enable_service_links=True,
                ephemeral_containers=[None],
                host_aliases=[None],
                host_ipc=True,
                host_network=True,
                host_pid=True,
                hostname="0",
                image_pull_secrets=[None],
                init_containers=[None],
                node_name="0",
                node_selector={"key": "0"},
                overhead={"key": None},
                preemption_policy="0",
                priority=56,
                priority_class_name="0",
                readiness_gates=[None],
                restart_policy="0",
                runtime_class_name="0",
                scheduler_name="0",
                security_context=None,
                service_account="0",
                service_account_name="0",
                share_process_namespace=True,
                subdomain="0",
                termination_grace_period_seconds=56,
                tolerations=[None],
                topology_spread_constraints=[None],
                volumes=[None],
            )
        else:
            return V1beta1PodSpec()

    def testV1beta1PodSpec(self):
        """Test V1beta1PodSpec"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
KServe

Python SDK for KServe  # noqa: E501

OpenAPI spec version: v0.1

Generated by: https://github.com/swagger-api/swagger-codegen.git
"""


from __future__ import absolute_import

import unittest

import kserve
from kserve.models.knative_addressable import KnativeAddressable  # noqa: E501
from kserve.rest import ApiException


class TestKnativeAddressable(unittest.TestCase):
    """KnativeAddressable unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def testKnativeAddressable(self):
        """Test KnativeAddressable"""
        # FIXME: construct object with mandatory attributes with example values
        # model = kserve.models.knative_addressable.KnativeAddressable()  # noqa: E501
        pass


if __name__ == "__main__":
    unittest.main()



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from socket import socket
import sys
from typing import List, Optional

import fastapi
import uvicorn
from fastapi import Request, Response
from fastapi.routing import APIRouter
from prometheus_client import REGISTRY, exposition
from timing_asgi import TimingClient, TimingMiddleware
from timing_asgi.integrations import StarletteScopeToName
from uvicorn.importer import import_from_string, ImportFromStringError

from kserve.errors import (
    InferenceError,
    InvalidInput,
    ModelNotFound,
    ModelNotReady,
    ServerNotLive,
    ServerNotReady,
    UnsupportedProtocol,
    generic_exception_handler,
    inference_error_handler,
    invalid_input_handler,
    model_not_found_handler,
    model_not_ready_handler,
    not_implemented_error_handler,
    server_not_live_handler,
    server_not_ready_handler,
    unsupported_protocol_error_handler,
)
from kserve.logging import trace_logger, logger
from kserve.protocol.dataplane import DataPlane

from .v1_endpoints import register_v1_endpoints
from .v2_endpoints import register_v2_endpoints
from ..model_repository_extension import ModelRepositoryExtension


async def metrics_handler(request: Request) -> Response:
    encoder, content_type = exposition.choose_encoder(request.headers.get("accept"))
    return Response(content=encoder(REGISTRY), headers={"content-type": content_type})


class PrintTimings(TimingClient):
    def timing(self, metric_name, timing, tags):
        trace_logger.info(f"{metric_name}: {timing} {tags}")


class RESTServer:
    def __init__(
        self,
        app: str,
        data_plane: DataPlane,
        model_repository_extension: ModelRepositoryExtension,
        http_port: int,
        access_log_format: Optional[str] = None,
        workers: int = 1,
        grace_period: int = 30,
    ):
        self.dataplane = data_plane
        self.model_repository_extension = model_repository_extension
        self.access_log_format = access_log_format
        self.config = uvicorn.Config(
            app,
            host="0.0.0.0",
            port=http_port,
            workers=workers,
            # By setting log_config to None we tell Uvicorn not to configure logging as it is already
            # configured by kserve.
            log_config=None,
            timeout_graceful_shutdown=grace_period,
            loop="asyncio",
        )
        self._server = uvicorn.Server(self.config)

    def _register_endpoints(self, app: fastapi.FastAPI):
        root_router = APIRouter()
        root_router.add_api_route(r"/", self.dataplane.live)
        root_router.add_api_route(r"/metrics", metrics_handler, methods=["GET"])
        app.include_router(root_router)
        register_v1_endpoints(app, self.dataplane, self.model_repository_extension)
        register_v2_endpoints(app, self.dataplane, self.model_repository_extension)
        # Register OpenAI endpoints if any of the models in the registry implement the OpenAI interface
        # This adds /openai/v1/completions and /openai/v1/chat/completions routes to the
        # REST server.
        try:
            from kserve.protocol.rest.openai.config import (
                maybe_register_openai_endpoints,
            )

            maybe_register_openai_endpoints(app, self.dataplane.model_registry)
            logger.info("OpenAI endpoints registered")
        except ImportError:
            logger.info("OpenAI endpoints not registered")

    def _add_exception_handlers(self, app: fastapi.FastAPI):
        app.add_exception_handler(InvalidInput, invalid_input_handler)
        app.add_exception_handler(InferenceError, inference_error_handler)
        app.add_exception_handler(ModelNotFound, model_not_found_handler)
        app.add_exception_handler(ModelNotReady, model_not_ready_handler)
        app.add_exception_handler(NotImplementedError, not_implemented_error_handler)
        app.add_exception_handler(
            UnsupportedProtocol, unsupported_protocol_error_handler
        )
        app.add_exception_handler(ServerNotLive, server_not_live_handler)
        app.add_exception_handler(ServerNotReady, server_not_ready_handler)
        app.add_exception_handler(Exception, generic_exception_handler)

    def _add_middlewares(self, app: fastapi.FastAPI):
        app.add_middleware(
            TimingMiddleware,
            client=PrintTimings(),
            metric_namer=StarletteScopeToName(prefix="kserve.io", starlette_app=app),
        )

        # More context in https://github.com/encode/uvicorn/pull/947
        # At the time of writing the ASGI specs are not clear when it comes
        # to change the access log format, and hence the Uvicorn upstream devs
        # chose to create a custom middleware for this.
        # The allowed log format is specified in https://github.com/Kludex/asgi-logger#usage
        if self.access_log_format:
            from asgi_logger import AccessLoggerMiddleware

            # As indicated by the asgi-logger docs, we need to clear/unset
            # any setting for uvicorn.access to avoid log duplicates.
            logging.getLogger("uvicorn.access").handlers = []
            app.add_middleware(AccessLoggerMiddleware, format=self.access_log_format)
            # The asgi-logger settings don't set propagate to False,
            # so we get duplicates if we don't set it explicitly.
            logging.getLogger("access").propagate = False

    def create_application(self):
        try:
            app = import_from_string("kserve.model_server:app")
        except ImportFromStringError as exc:
            logger.error("Error loading ASGI app. %s", exc)
            sys.exit(1)
        self._add_middlewares(app)
        self._register_endpoints(app)
        self._add_exception_handlers(app)

    async def start(self):
        """Starts the server without configuring the event loop."""
        self.create_application()
        logger.info("Starting uvicorn with %s workers", self.config.workers)
        await self._server.serve()

    def run(self, sockets: Optional[List[socket]] = None):
        """Configures event loop and starts the server."""
        self.create_application()
        self._server.run(sockets=sockets)



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
KServe

Python SDK for KServe  # noqa: E501

The version of the OpenAPI document: v0.1
Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1beta1_inference_service_spec import (
    V1beta1InferenceServiceSpec,
)  # noqa: E501
from kserve.rest import ApiException


class TestV1beta1InferenceServiceSpec(unittest.TestCase):
    """V1beta1InferenceServiceSpec unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1beta1InferenceServiceSpec
        include_option is a boolean, when False only required
        params are included, when True both required and
        optional params are included"""
        # model = kserve.models.v1beta1_inference_service_spec.V1beta1InferenceServiceSpec()  # noqa: E501
        if include_optional:
            return V1beta1InferenceServiceSpec(
                explainer=kserve.models.v1beta1_explainer_spec.V1beta1ExplainerSpec(
                    active_deadline_seconds=56,
                    affinity=None,
                    art=kserve.models.v1beta1_art_explainer_spec.V1beta1ARTExplainerSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        type="0",
                        working_dir="0",
                    ),
                    automount_service_account_token=True,
                    batcher=kserve.models.v1beta1_batcher.V1beta1Batcher(
                        max_batch_size=56,
                        max_latency=56,
                        timeout=56,
                    ),
                    canary_traffic_percent=56,
                    container_concurrency=56,
                    containers=[None],
                    dns_config=None,
                    dns_policy="0",
                    enable_service_links=True,
                    ephemeral_containers=[None],
                    host_aliases=[None],
                    host_ipc=True,
                    host_network=True,
                    host_pid=True,
                    hostname="0",
                    image_pull_secrets=[None],
                    init_containers=[None],
                    logger=kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                        mode="0",
                        url="0",
                    ),
                    max_replicas=56,
                    min_replicas=56,
                    node_name="0",
                    node_selector={"key": "0"},
                    overhead={"key": None},
                    preemption_policy="0",
                    priority=56,
                    priority_class_name="0",
                    readiness_gates=[None],
                    restart_policy="0",
                    runtime_class_name="0",
                    scheduler_name="0",
                    security_context=None,
                    service_account="0",
                    service_account_name="0",
                    share_process_namespace=True,
                    subdomain="0",
                    termination_grace_period_seconds=56,
                    timeout=56,
                    tolerations=[None],
                    topology_spread_constraints=[None],
                    volumes=[None],
                ),
                predictor=kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                    active_deadline_seconds=56,
                    affinity=None,
                    automount_service_account_token=True,
                    batcher=kserve.models.v1beta1_batcher.V1beta1Batcher(
                        max_batch_size=56,
                        max_latency=56,
                        timeout=56,
                    ),
                    canary_traffic_percent=56,
                    container_concurrency=56,
                    containers=[None],
                    dns_config=None,
                    dns_policy="0",
                    enable_service_links=True,
                    ephemeral_containers=[None],
                    host_aliases=[None],
                    host_ipc=True,
                    host_network=True,
                    host_pid=True,
                    hostname="0",
                    image_pull_secrets=[None],
                    init_containers=[None],
                    logger=kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                        mode="0",
                        url="0",
                    ),
                    max_replicas=56,
                    min_replicas=56,
                    node_name="0",
                    node_selector={"key": "0"},
                    onnx=kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                        args=["0"],
                        command=["0"],
                        env=[None],
                        env_from=[None],
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        ports=[None],
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        volume_devices=[None],
                        volume_mounts=[None],
                        working_dir="0",
                    ),
                    overhead={"key": None},
                    preemption_policy="0",
                    priority=56,
                    priority_class_name="0",
                    pytorch=kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                    readiness_gates=[None],
                    restart_policy="0",
                    runtime_class_name="0",
                    scheduler_name="0",
                    security_context=None,
                    service_account="0",
                    service_account_name="0",
                    share_process_namespace=True,
                    sklearn=kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                    subdomain="0",
                    tensorflow=kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                    termination_grace_period_seconds=56,
                    timeout=56,
                    tolerations=[None],
                    topology_spread_constraints=[None],
                    triton=kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                    volumes=[None],
                    xgboost=kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                ),
                transformer=kserve.models.v1beta1_transformer_spec.V1beta1TransformerSpec(
                    active_deadline_seconds=56,
                    affinity=None,
                    automount_service_account_token=True,
                    batcher=kserve.models.v1beta1_batcher.V1beta1Batcher(
                        max_batch_size=56,
                        max_latency=56,
                        timeout=56,
                    ),
                    canary_traffic_percent=56,
                    container_concurrency=56,
                    containers=[None],
                    dns_config=None,
                    dns_policy="0",
                    enable_service_links=True,
                    ephemeral_containers=[None],
                    host_aliases=[None],
                    host_ipc=True,
                    host_network=True,
                    host_pid=True,
                    hostname="0",
                    image_pull_secrets=[None],
                    init_containers=[None],
                    logger=kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                        mode="0",
                        url="0",
                    ),
                    max_replicas=56,
                    min_replicas=56,
                    node_name="0",
                    node_selector={"key": "0"},
                    overhead={"key": None},
                    preemption_policy="0",
                    priority=56,
                    priority_class_name="0",
                    readiness_gates=[None],
                    restart_policy="0",
                    runtime_class_name="0",
                    scheduler_name="0",
                    security_context=None,
                    service_account="0",
                    service_account_name="0",
                    share_process_namespace=True,
                    subdomain="0",
                    termination_grace_period_seconds=56,
                    timeout=56,
                    tolerations=[None],
                    topology_spread_constraints=[None],
                    volumes=[None],
                ),
            )
        else:
            return V1beta1InferenceServiceSpec(
                predictor=kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                    active_deadline_seconds=56,
                    affinity=None,
                    automount_service_account_token=True,
                    batcher=kserve.models.v1beta1_batcher.V1beta1Batcher(
                        max_batch_size=56,
                        max_latency=56,
                        timeout=56,
                    ),
                    canary_traffic_percent=56,
                    container_concurrency=56,
                    containers=[None],
                    dns_config=None,
                    dns_policy="0",
                    enable_service_links=True,
                    ephemeral_containers=[None],
                    host_aliases=[None],
                    host_ipc=True,
                    host_network=True,
                    host_pid=True,
                    hostname="0",
                    image_pull_secrets=[None],
                    init_containers=[None],
                    logger=kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                        mode="0",
                        url="0",
                    ),
                    max_replicas=56,
                    min_replicas=56,
                    node_name="0",
                    node_selector={"key": "0"},
                    onnx=kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                        args=["0"],
                        command=["0"],
                        env=[None],
                        env_from=[None],
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        ports=[None],
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        volume_devices=[None],
                        volume_mounts=[None],
                        working_dir="0",
                    ),
                    overhead={"key": None},
                    preemption_policy="0",
                    priority=56,
                    priority_class_name="0",
                    pytorch=kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                    readiness_gates=[None],
                    restart_policy="0",
                    runtime_class_name="0",
                    scheduler_name="0",
                    security_context=None,
                    service_account="0",
                    service_account_name="0",
                    share_process_namespace=True,
                    sklearn=kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                    subdomain="0",
                    tensorflow=kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                    termination_grace_period_seconds=56,
                    timeout=56,
                    tolerations=[None],
                    topology_spread_constraints=[None],
                    triton=kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                    volumes=[None],
                    xgboost=kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                        image="0",
                        image_pull_policy="0",
                        lifecycle=None,
                        liveness_probe=None,
                        name="0",
                        readiness_probe=None,
                        resources=None,
                        runtime_version="0",
                        security_context=None,
                        startup_probe=None,
                        stdin=True,
                        stdin_once=True,
                        storage_uri="0",
                        termination_message_path="0",
                        termination_message_policy="0",
                        tty=True,
                        working_dir="0",
                    ),
                ),
            )

    def testV1beta1InferenceServiceSpec(self):
        """Test V1beta1InferenceServiceSpec"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == "__main__":
    unittest.main()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2022 The KServe Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class ServerLiveRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ServerLiveRequest = ServerLiveRequest

@typing.final
class ServerLiveResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LIVE_FIELD_NUMBER: builtins.int
    live: builtins.bool
    """True if the inference server is live, false if not live."""
    def __init__(
        self,
        *,
        live: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["live", b"live"]) -> None: ...

global___ServerLiveResponse = ServerLiveResponse

@typing.final
class ServerReadyRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ServerReadyRequest = ServerReadyRequest

@typing.final
class ServerReadyResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    READY_FIELD_NUMBER: builtins.int
    ready: builtins.bool
    """True if the inference server is ready, false if not ready."""
    def __init__(
        self,
        *,
        ready: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["ready", b"ready"]) -> None: ...

global___ServerReadyResponse = ServerReadyResponse

@typing.final
class ModelReadyRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name of the model to check for readiness."""
    version: builtins.str
    """The version of the model to check for readiness. If not given the
    server will choose a version based on the model and internal policy.
    """
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        version: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name", "version", b"version"]) -> None: ...

global___ModelReadyRequest = ModelReadyRequest

@typing.final
class ModelReadyResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    READY_FIELD_NUMBER: builtins.int
    ready: builtins.bool
    """True if the model is ready, false if not ready."""
    def __init__(
        self,
        *,
        ready: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["ready", b"ready"]) -> None: ...

global___ModelReadyResponse = ModelReadyResponse

@typing.final
class ServerMetadataRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ServerMetadataRequest = ServerMetadataRequest

@typing.final
class ServerMetadataResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    EXTENSIONS_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The server name."""
    version: builtins.str
    """The server version."""
    @property
    def extensions(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The extensions supported by the server."""

    def __init__(
        self,
        *,
        name: builtins.str = ...,
        version: builtins.str = ...,
        extensions: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["extensions", b"extensions", "name", b"name", "version", b"version"]) -> None: ...

global___ServerMetadataResponse = ServerMetadataResponse

@typing.final
class ModelMetadataRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The name of the model."""
    version: builtins.str
    """The version of the model to check for readiness. If not given the
    server will choose a version based on the model and internal policy.
    """
    def __init__(
        self,
        *,
        name: builtins.str = ...,
        version: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["name", b"name", "version", b"version"]) -> None: ...

global___ModelMetadataRequest = ModelMetadataRequest

@typing.final
class ModelMetadataResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class TensorMetadata(google.protobuf.message.Message):
        """Metadata for a tensor."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        NAME_FIELD_NUMBER: builtins.int
        DATATYPE_FIELD_NUMBER: builtins.int
        SHAPE_FIELD_NUMBER: builtins.int
        name: builtins.str
        """The tensor name."""
        datatype: builtins.str
        """The tensor data type."""
        @property
        def shape(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
            """The tensor shape. A variable-size dimension is represented
            by a -1 value.
            """

        def __init__(
            self,
            *,
            name: builtins.str = ...,
            datatype: builtins.str = ...,
            shape: collections.abc.Iterable[builtins.int] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["datatype", b"datatype", "name", b"name", "shape", b"shape"]) -> None: ...

    NAME_FIELD_NUMBER: builtins.int
    VERSIONS_FIELD_NUMBER: builtins.int
    PLATFORM_FIELD_NUMBER: builtins.int
    INPUTS_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    name: builtins.str
    """The model name."""
    platform: builtins.str
    """The model's platform. See Platforms."""
    @property
    def versions(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The versions of the model available on the server."""

    @property
    def inputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ModelMetadataResponse.TensorMetadata]:
        """The model's inputs."""

    @property
    def outputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ModelMetadataResponse.TensorMetadata]:
        """The model's outputs."""

    def __init__(
        self,
        *,
        name: builtins.str = ...,
        versions: collections.abc.Iterable[builtins.str] | None = ...,
        platform: builtins.str = ...,
        inputs: collections.abc.Iterable[global___ModelMetadataResponse.TensorMetadata] | None = ...,
        outputs: collections.abc.Iterable[global___ModelMetadataResponse.TensorMetadata] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["inputs", b"inputs", "name", b"name", "outputs", b"outputs", "platform", b"platform", "versions", b"versions"]) -> None: ...

global___ModelMetadataResponse = ModelMetadataResponse

@typing.final
class ModelInferRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class InferInputTensor(google.protobuf.message.Message):
        """An input tensor for an inference request."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class ParametersEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___InferParameter: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___InferParameter | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        NAME_FIELD_NUMBER: builtins.int
        DATATYPE_FIELD_NUMBER: builtins.int
        SHAPE_FIELD_NUMBER: builtins.int
        PARAMETERS_FIELD_NUMBER: builtins.int
        CONTENTS_FIELD_NUMBER: builtins.int
        name: builtins.str
        """The tensor name."""
        datatype: builtins.str
        """The tensor data type."""
        @property
        def shape(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
            """The tensor shape."""

        @property
        def parameters(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___InferParameter]:
            """Optional inference input tensor parameters."""

        @property
        def contents(self) -> global___InferTensorContents:
            """The tensor contents using a data-type format. This field must
            not be specified if "raw" tensor contents are being used for
            the inference request.
            """

        def __init__(
            self,
            *,
            name: builtins.str = ...,
            datatype: builtins.str = ...,
            shape: collections.abc.Iterable[builtins.int] | None = ...,
            parameters: collections.abc.Mapping[builtins.str, global___InferParameter] | None = ...,
            contents: global___InferTensorContents | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["contents", b"contents"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["contents", b"contents", "datatype", b"datatype", "name", b"name", "parameters", b"parameters", "shape", b"shape"]) -> None: ...

    @typing.final
    class InferRequestedOutputTensor(google.protobuf.message.Message):
        """An output tensor requested for an inference request."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class ParametersEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___InferParameter: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___InferParameter | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        NAME_FIELD_NUMBER: builtins.int
        PARAMETERS_FIELD_NUMBER: builtins.int
        name: builtins.str
        """The tensor name."""
        @property
        def parameters(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___InferParameter]:
            """Optional requested output tensor parameters."""

        def __init__(
            self,
            *,
            name: builtins.str = ...,
            parameters: collections.abc.Mapping[builtins.str, global___InferParameter] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["name", b"name", "parameters", b"parameters"]) -> None: ...

    @typing.final
    class ParametersEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___InferParameter: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___InferParameter | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    MODEL_NAME_FIELD_NUMBER: builtins.int
    MODEL_VERSION_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    PARAMETERS_FIELD_NUMBER: builtins.int
    INPUTS_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    RAW_INPUT_CONTENTS_FIELD_NUMBER: builtins.int
    model_name: builtins.str
    """The name of the model to use for inferencing."""
    model_version: builtins.str
    """The version of the model to use for inference. If not given the
    server will choose a version based on the model and internal policy.
    """
    id: builtins.str
    """Optional identifier for the request. If specified will be
    returned in the response.
    """
    @property
    def parameters(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___InferParameter]:
        """Optional inference parameters."""

    @property
    def inputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ModelInferRequest.InferInputTensor]:
        """The input tensors for the inference."""

    @property
    def outputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ModelInferRequest.InferRequestedOutputTensor]:
        """The requested output tensors for the inference. Optional, if not
        specified all outputs produced by the model will be returned.
        """

    @property
    def raw_input_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]:
        """The data contained in an input tensor can be represented in "raw"
        bytes form or in the repeated type that matches the tensor's data
        type. To use the raw representation 'raw_input_contents' must be
        initialized with data for each tensor in the same order as
        'inputs'. For each tensor, the size of this content must match
        what is expected by the tensor's shape and data type. The raw
        data must be the flattened, one-dimensional, row-major order of
        the tensor elements without any stride or padding between the
        elements. Note that the FP16 and BF16 data types must be represented as
        raw content as there is no specific data type for a 16-bit float type.

        If this field is specified then InferInputTensor::contents must
        not be specified for any input tensor.
        """

    def __init__(
        self,
        *,
        model_name: builtins.str = ...,
        model_version: builtins.str = ...,
        id: builtins.str = ...,
        parameters: collections.abc.Mapping[builtins.str, global___InferParameter] | None = ...,
        inputs: collections.abc.Iterable[global___ModelInferRequest.InferInputTensor] | None = ...,
        outputs: collections.abc.Iterable[global___ModelInferRequest.InferRequestedOutputTensor] | None = ...,
        raw_input_contents: collections.abc.Iterable[builtins.bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["id", b"id", "inputs", b"inputs", "model_name", b"model_name", "model_version", b"model_version", "outputs", b"outputs", "parameters", b"parameters", "raw_input_contents", b"raw_input_contents"]) -> None: ...

global___ModelInferRequest = ModelInferRequest

@typing.final
class ModelInferResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class InferOutputTensor(google.protobuf.message.Message):
        """An output tensor returned for an inference request."""

        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class ParametersEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___InferParameter: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___InferParameter | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        NAME_FIELD_NUMBER: builtins.int
        DATATYPE_FIELD_NUMBER: builtins.int
        SHAPE_FIELD_NUMBER: builtins.int
        PARAMETERS_FIELD_NUMBER: builtins.int
        CONTENTS_FIELD_NUMBER: builtins.int
        name: builtins.str
        """The tensor name."""
        datatype: builtins.str
        """The tensor data type."""
        @property
        def shape(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
            """The tensor shape."""

        @property
        def parameters(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___InferParameter]:
            """Optional output tensor parameters."""

        @property
        def contents(self) -> global___InferTensorContents:
            """The tensor contents using a data-type format. This field must
            not be specified if "raw" tensor contents are being used for
            the inference response.
            """

        def __init__(
            self,
            *,
            name: builtins.str = ...,
            datatype: builtins.str = ...,
            shape: collections.abc.Iterable[builtins.int] | None = ...,
            parameters: collections.abc.Mapping[builtins.str, global___InferParameter] | None = ...,
            contents: global___InferTensorContents | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["contents", b"contents"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["contents", b"contents", "datatype", b"datatype", "name", b"name", "parameters", b"parameters", "shape", b"shape"]) -> None: ...

    @typing.final
    class ParametersEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___InferParameter: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___InferParameter | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    MODEL_NAME_FIELD_NUMBER: builtins.int
    MODEL_VERSION_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    PARAMETERS_FIELD_NUMBER: builtins.int
    OUTPUTS_FIELD_NUMBER: builtins.int
    RAW_OUTPUT_CONTENTS_FIELD_NUMBER: builtins.int
    model_name: builtins.str
    """The name of the model used for inference."""
    model_version: builtins.str
    """The version of the model used for inference."""
    id: builtins.str
    """The id of the inference request if one was specified."""
    @property
    def parameters(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___InferParameter]:
        """Optional inference response parameters."""

    @property
    def outputs(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ModelInferResponse.InferOutputTensor]:
        """The output tensors holding inference results."""

    @property
    def raw_output_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]:
        """The data contained in an output tensor can be represented in
        "raw" bytes form or in the repeated type that matches the
        tensor's data type. To use the raw representation 'raw_output_contents'
        must be initialized with data for each tensor in the same order as
        'outputs'. For each tensor, the size of this content must match
        what is expected by the tensor's shape and data type. The raw
        data must be the flattened, one-dimensional, row-major order of
        the tensor elements without any stride or padding between the
        elements. Note that the FP16 and BF16 data types must be represented as
        raw content as there is no specific data type for a 16-bit float type.

        If this field is specified then InferOutputTensor::contents must
        not be specified for any output tensor.
        """

    def __init__(
        self,
        *,
        model_name: builtins.str = ...,
        model_version: builtins.str = ...,
        id: builtins.str = ...,
        parameters: collections.abc.Mapping[builtins.str, global___InferParameter] | None = ...,
        outputs: collections.abc.Iterable[global___ModelInferResponse.InferOutputTensor] | None = ...,
        raw_output_contents: collections.abc.Iterable[builtins.bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["id", b"id", "model_name", b"model_name", "model_version", b"model_version", "outputs", b"outputs", "parameters", b"parameters", "raw_output_contents", b"raw_output_contents"]) -> None: ...

global___ModelInferResponse = ModelInferResponse

@typing.final
class InferParameter(google.protobuf.message.Message):
    """An inference parameter value. The Parameters message describes a 
    “name”/”value” pair, where the “name” is the name of the parameter
    and the “value” is a boolean, integer, or string corresponding to 
    the parameter.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BOOL_PARAM_FIELD_NUMBER: builtins.int
    INT64_PARAM_FIELD_NUMBER: builtins.int
    STRING_PARAM_FIELD_NUMBER: builtins.int
    bool_param: builtins.bool
    """A boolean parameter value."""
    int64_param: builtins.int
    """An int64 parameter value."""
    string_param: builtins.str
    """A string parameter value."""
    def __init__(
        self,
        *,
        bool_param: builtins.bool = ...,
        int64_param: builtins.int = ...,
        string_param: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bool_param", b"bool_param", "int64_param", b"int64_param", "parameter_choice", b"parameter_choice", "string_param", b"string_param"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bool_param", b"bool_param", "int64_param", b"int64_param", "parameter_choice", b"parameter_choice", "string_param", b"string_param"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["parameter_choice", b"parameter_choice"]) -> typing.Literal["bool_param", "int64_param", "string_param"] | None: ...

global___InferParameter = InferParameter

@typing.final
class InferTensorContents(google.protobuf.message.Message):
    """The data contained in a tensor represented by the repeated type
    that matches the tensor's data type. Protobuf oneof is not used
    because oneofs cannot contain repeated fields.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BOOL_CONTENTS_FIELD_NUMBER: builtins.int
    INT_CONTENTS_FIELD_NUMBER: builtins.int
    INT64_CONTENTS_FIELD_NUMBER: builtins.int
    UINT_CONTENTS_FIELD_NUMBER: builtins.int
    UINT64_CONTENTS_FIELD_NUMBER: builtins.int
    FP32_CONTENTS_FIELD_NUMBER: builtins.int
    FP64_CONTENTS_FIELD_NUMBER: builtins.int
    BYTES_CONTENTS_FIELD_NUMBER: builtins.int
    @property
    def bool_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bool]:
        """Representation for BOOL data type. The size must match what is
        expected by the tensor's shape. The contents must be the flattened,
        one-dimensional, row-major order of the tensor elements.
        """

    @property
    def int_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Representation for INT8, INT16, and INT32 data types. The size
        must match what is expected by the tensor's shape. The contents
        must be the flattened, one-dimensional, row-major order of the
        tensor elements.
        """

    @property
    def int64_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Representation for INT64 data types. The size must match what
        is expected by the tensor's shape. The contents must be the
        flattened, one-dimensional, row-major order of the tensor elements.
        """

    @property
    def uint_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Representation for UINT8, UINT16, and UINT32 data types. The size
        must match what is expected by the tensor's shape. The contents
        must be the flattened, one-dimensional, row-major order of the
        tensor elements.
        """

    @property
    def uint64_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Representation for UINT64 data types. The size must match what
        is expected by the tensor's shape. The contents must be the
        flattened, one-dimensional, row-major order of the tensor elements.
        """

    @property
    def fp32_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """Representation for FP32 data type. The size must match what is
        expected by the tensor's shape. The contents must be the flattened,
        one-dimensional, row-major order of the tensor elements.
        """

    @property
    def fp64_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """Representation for FP64 data type. The size must match what is
        expected by the tensor's shape. The contents must be the flattened,
        one-dimensional, row-major order of the tensor elements.
        """

    @property
    def bytes_contents(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]:
        """Representation for BYTES data type. The size must match what is
        expected by the tensor's shape. The contents must be the flattened,
        one-dimensional, row-major order of the tensor elements.
        """

    def __init__(
        self,
        *,
        bool_contents: collections.abc.Iterable[builtins.bool] | None = ...,
        int_contents: collections.abc.Iterable[builtins.int] | None = ...,
        int64_contents: collections.abc.Iterable[builtins.int] | None = ...,
        uint_contents: collections.abc.Iterable[builtins.int] | None = ...,
        uint64_contents: collections.abc.Iterable[builtins.int] | None = ...,
        fp32_contents: collections.abc.Iterable[builtins.float] | None = ...,
        fp64_contents: collections.abc.Iterable[builtins.float] | None = ...,
        bytes_contents: collections.abc.Iterable[builtins.bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["bool_contents", b"bool_contents", "bytes_contents", b"bytes_contents", "fp32_contents", b"fp32_contents", "fp64_contents", b"fp64_contents", "int64_contents", b"int64_contents", "int_contents", b"int_contents", "uint64_contents", b"uint64_contents", "uint_contents", b"uint_contents"]) -> None: ...

global___InferTensorContents = InferTensorContents

@typing.final
class RepositoryModelLoadRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODEL_NAME_FIELD_NUMBER: builtins.int
    model_name: builtins.str
    """The name of the model to load, or reload."""
    def __init__(
        self,
        *,
        model_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["model_name", b"model_name"]) -> None: ...

global___RepositoryModelLoadRequest = RepositoryModelLoadRequest

@typing.final
class RepositoryModelLoadResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODEL_NAME_FIELD_NUMBER: builtins.int
    ISLOADED_FIELD_NUMBER: builtins.int
    model_name: builtins.str
    """The name of the model trying to load or reload."""
    isLoaded: builtins.bool
    """boolean parameter to indicate whether model is loaded or not"""
    def __init__(
        self,
        *,
        model_name: builtins.str = ...,
        isLoaded: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["isLoaded", b"isLoaded", "model_name", b"model_name"]) -> None: ...

global___RepositoryModelLoadResponse = RepositoryModelLoadResponse

@typing.final
class RepositoryModelUnloadRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODEL_NAME_FIELD_NUMBER: builtins.int
    model_name: builtins.str
    """The name of the model to unload."""
    def __init__(
        self,
        *,
        model_name: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["model_name", b"model_name"]) -> None: ...

global___RepositoryModelUnloadRequest = RepositoryModelUnloadRequest

@typing.final
class RepositoryModelUnloadResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MODEL_NAME_FIELD_NUMBER: builtins.int
    ISUNLOADED_FIELD_NUMBER: builtins.int
    model_name: builtins.str
    """The name of the model trying to load or reload."""
    isUnloaded: builtins.bool
    """boolean parameter to indicate whether model is unloaded or not"""
    def __init__(
        self,
        *,
        model_name: builtins.str = ...,
        isUnloaded: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["isUnloaded", b"isUnloaded", "model_name", b"model_name"]) -> None: ...

global___RepositoryModelUnloadResponse = RepositoryModelUnloadResponse
